from math import *
from euclid import *
from omega import *
from cyclops import *
from CoordinateCalculator import CoordinateCalculator
import time

# ------------------------
# General initializations
# ------------------------
# Clipping doesn't exist in my build of omegalib
clippingOn = True

mainMenu = MenuManager.createAndInitialize()
getSceneManager().setBackgroundColor(Color("#101010"))

# Required demon magic needed to create widgets
ui = UiModule.createAndInitialize()
wf = ui.getWidgetFactory()
uiRoot = ui.getUi()

# Code convenience variables
maxMultChildren = 14 # Gliese 667 hits this limit
tab = "    " # ...no escape characters

cam = getDefaultCamera()
cam.setPosition(Vector3(0.0, 0.0, 15.0))
cam.setPitchYawRoll(Vector3(0.0, 0.0, 0.0))
cam.getController().setSpeed(10)
setNearFarZ(0.001, 1000.0)

# ----------------------------------------------------------
# Ridiculously gigantic object specific XML parsing classes
# ----------------------------------------------------------
class System:
	def __init__(self, inData):
		# childType can have values 1, 2, or 3 and associate with "Binary", "Star", or "Planet"
		# childId is the index within allBinaries[], allStars[], or allPlanets[]
		self.children = [] # [childType, childId]
		
		# Data specific to System instances from XML
		self.name = [] # String array (some number of names >= 1)
		self.distance = [0.0, 0.0, 0.0] # [value, error minus, error plus]
		self.rightAscension = ""
		self.declination = ""
		
		data = inData.split("\n")
		for line in data:
			if line:
				endOpen = 1
				while line[endOpen] != ">": endOpen += 1
				startClose = endOpen+1
				while line[startClose] != "<": startClose += 1
				
				# <tagOpen>tagValue</foo>
				tagOpen = line[1:endOpen]
				tagValue = line[endOpen+1:startClose]
				spaceSplit = tagOpen.split(" ")
				
				# If there are no spaces then there are no error attributes
				if (len(spaceSplit) == 1):
					if tagOpen == "name": self.name.append(tagValue)
					elif tagOpen == "distance": self.distance[0] = float(tagValue)
					elif tagOpen == "rightascension": self.rightAscension = tagValue
					elif tagOpen == "declination": self.declination = tagValue
					else: print "Not implemented in System: "+line
				
				# If there are spaces then there are error attributes
				else:
					if spaceSplit[0] == "distance":
						self.distance[0] = float(tagValue)
						self.distance[1] = float(spaceSplit[1][12:-1]) # error minus
						self.distance[2] = float(spaceSplit[2][11:-1]) # error plus
						if spaceSplit[1][:10] != "errorminus" or spaceSplit[2][:9] != "errorplus": print "...something went wrong reading errors in System"
					else: print "Not implemented in System: "+line

class Binary:
	def __init__(self, inData):
		# parentType can have values 0 or 1 and associate with "System" or "Binary"
		# parentId is the index within allSystems[] or allBinaries[]
		self.parent = [-1, -1] # [parentType, parentId]
		
		# childType can have values 1, 2, or 3 and associate with "Binary", "Star", or "Planet"
		# childId is the index within allBinaries[], allStars[], or allPlanets[]
		self.children = [] # [childType, childId]
		
		# Data specific to Binary instances from XML
		self.semiMajorAxis = [0.0, 0.0, 0.0] # [value, error minus, error plus]
		self.eccentricity = [0.0, 0.0, 0.0] # [value, error minus, error plus]
		self.inclination = [0.0, 0.0, 0.0] # [value, error minus, error plus]
		self.period = [0.0, 0.0, 0.0] # [value, error minus, error plus]
		
		data = inData.split("\n")
		for line in data:
			if line:
				endOpen = 1
				while line[endOpen] != ">": endOpen += 1
				startClose = endOpen+1
				while line[startClose] != "<": startClose += 1
				
				# <tagOpen>tagValue</foo>
				tagOpen = line[1:endOpen]
				tagValue = line[endOpen+1:startClose]
				spaceSplit = tagOpen.split(" ")
				
				# If there are no spaces then there are no error attributes
				if (len(spaceSplit) == 1):
					if tagOpen == "semimajoraxis": self.semiMajorAxis[0] = float(tagValue)
					elif tagOpen == "eccentricity": self.eccentricity[0] = float(tagValue)
					elif tagOpen == "inclination": self.inclination[0] = float(tagValue)
					elif tagOpen == "period": self.period[0] = float(tagValue)
					elif tagOpen == "longitude": continue # Ignore for now...
					elif tagOpen == "periastron": continue # Ignore for now...
					elif tagOpen == "ascendingnode": continue # Ignore for now...
					else: print "Not implemented in Binary: "+line
				
				# If there are spaces then there are error attributes
				else:
					if spaceSplit[0] == "semimajoraxis":
						self.semiMajorAxis[0] = float(tagValue)
						self.semiMajorAxis[1] = float(spaceSplit[1][12:-1]) # error minus
						self.semiMajorAxis[2] = float(spaceSplit[2][11:-1]) # error plus
						if spaceSplit[1][:10] != "errorminus" or spaceSplit[2][:9] != "errorplus": print "...something went wrong reading errors in Binary"
					elif spaceSplit[0] == "eccentricity":
						self.eccentricity[0] = float(tagValue)
						self.eccentricity[1] = float(spaceSplit[1][12:-1]) # error minus
						self.eccentricity[2] = float(spaceSplit[2][11:-1]) # error plus
						if spaceSplit[1][:10] != "errorminus" or spaceSplit[2][:9] != "errorplus": print "...something went wrong reading errors in Binary"
					elif spaceSplit[0] == "inclination":
						self.inclination[0] = float(tagValue)
						self.inclination[1] = float(spaceSplit[1][12:-1]) # error minus
						self.inclination[2] = float(spaceSplit[2][11:-1]) # error plus
						if spaceSplit[1][:10] != "errorminus" or spaceSplit[2][:9] != "errorplus": print "...something went wrong reading errors in Binary"
					elif spaceSplit[0] == "period":
						self.period[0] = float(tagValue)
						self.period[1] = float(spaceSplit[1][12:-1]) # error minus
						self.period[2] = float(spaceSplit[2][11:-1]) # error plus
						if spaceSplit[1][:10] != "errorminus" or spaceSplit[2][:9] != "errorplus": print "...something went wrong reading errors in Binary"
					elif spaceSplit[0] == "longitude": continue # Ignore for now...
					elif spaceSplit[0] == "periastron": continue # Ignore for now...
					elif spaceSplit[0] == "ascendingnode": continue # Ignore for now...
					else: print "Not implemented in Binary: "+line

class Star:
	def __init__(self, inData):
		# parentType can have values 0 or 1 and associate with "System" or "Binary"
		# parentId is the index within allSystems[] or allBinaries[]
		self.parent = [-1, -1] # [parentType, parentId]
		
		# childType has the value 3 and associates with "Planet"
		# childId is the index within allPlanets[]
		self.children = [] # [childType, childId]
		
		# Calculated later for the habitable zone
		self.habitableInner = 0.0
		self.habitableOuter = 0.0
		
		# Data specific to Star instances from XML
		self.name = [] # String array (some number of names >= 1)
		self.radius = [0.0, 0.0, 0.0] # [value, error minus, error plus]
		self.mass = [0.0, 0.0, 0.0] # [value, error minus, error plus]
		self.temperature = [0.0, 0.0, 0.0] # [value, error minus, error plus]
		self.magV = [0.0, 0.0, 0.0] # [value, error minus, error plus]
		self.metallicity = [0.0, 0.0, 0.0] # [value, error minus, error plus]
		self.age = [0.0, 0.0, 0.0] # [value, error minus, error plus]
		self.spectralType = ["", ""] # XML string, color string (assigned later)
		
		data = inData.split("\n")
		for line in data:
			if line:
				endOpen = 1
				while line[endOpen] != ">": endOpen += 1
				startClose = endOpen+1
				while line[startClose] != "<": startClose += 1
				
				# <tagOpen>tagValue</foo>
				tagOpen = line[1:endOpen]
				tagValue = line[endOpen+1:startClose]
				spaceSplit = tagOpen.split(" ")
				
				# If there are no spaces then there are no error attributes
				if (len(spaceSplit) == 1):
					if tagOpen == "name": self.name.append(tagValue)
					elif tagOpen == "radius": self.radius[0] = float(tagValue)
					elif tagOpen == "mass": self.mass[0] = float(tagValue)
					elif tagOpen == "temperature": self.temperature[0] = float(tagValue)
					elif tagOpen == "magV": self.magV[0] = float(tagValue)
					elif tagOpen == "metallicity": self.metallicity[0] = float(tagValue)
					elif tagOpen == "age": self.age[0] = float(tagValue)
					elif tagOpen == "spectraltype": self.spectralType[0] = tagValue
					else: print "Not implemented in Star: "+line
				
				# If there are spaces then there are error attributes
				else:
					if spaceSplit[0] == "radius":
						self.radius[0] = float(tagValue)
						self.radius[1] = float(spaceSplit[1][12:-1]) # error minus
						self.radius[2] = float(spaceSplit[2][11:-1]) # error plus
						if spaceSplit[1][:10] != "errorminus" or spaceSplit[2][:9] != "errorplus": print "...something went wrong reading errors in Star"
					elif spaceSplit[0] == "mass":
						self.mass[0] = float(tagValue)
						self.mass[1] = float(spaceSplit[1][12:-1]) # error minus
						self.mass[2] = float(spaceSplit[2][11:-1]) # error plus
						if spaceSplit[1][:10] != "errorminus" or spaceSplit[2][:9] != "errorplus": print "...something went wrong reading errors in Star"
					elif spaceSplit[0] == "temperature":
						self.temperature[0] = float(tagValue)
						self.temperature[1] = float(spaceSplit[1][12:-1]) # error minus
						self.temperature[2] = float(spaceSplit[2][11:-1]) # error plus
						if spaceSplit[1][:10] != "errorminus" or spaceSplit[2][:9] != "errorplus": print "...something went wrong reading errors in Star"
					elif spaceSplit[0] == "magV":
						self.magV[0] = float(tagValue)
						self.magV[1] = float(spaceSplit[1][12:-1]) # error minus
						self.magV[2] = float(spaceSplit[2][11:-1]) # error plus
						if spaceSplit[1][:10] != "errorminus" or spaceSplit[2][:9] != "errorplus": print "...something went wrong reading errors in Star"
					elif spaceSplit[0] == "metallicity":
						self.metallicity[0] = float(tagValue)
						self.metallicity[1] = float(spaceSplit[1][12:-1]) # error minus
						self.metallicity[2] = float(spaceSplit[2][11:-1]) # error plus
						if spaceSplit[1][:10] != "errorminus" or spaceSplit[2][:9] != "errorplus": print "...something went wrong reading errors in Star"
					elif spaceSplit[0] == "age":
						self.age[0] = float(tagValue)
						self.age[1] = float(spaceSplit[1][12:-1]) # error minus
						self.age[2] = float(spaceSplit[2][11:-1]) # error plus
						if spaceSplit[1][:10] != "errorminus" or spaceSplit[2][:9] != "errorplus": print "...something went wrong reading errors in Star"
					else: print "Not implemented in Star: "+line

class Planet:
	def __init__(self, inData):
		# parentType can have values 0, 1, or 2 and associate with "System", "Binary", or "Star"
		# parentId is the index within allSystems[], allBinaries[], or allStars
		self.parent = [-1, -1] # [parentType, parentId]
		
		# Data specific to Planet instances from XML
		self.name = [] # String array (some number of names >= 1)
		self.radius = [0.0, 0.0, 0.0] # [value, error minus, error plus]
		self.mass = [0.0, 0.0, 0.0] # [value, error minus, error plus]
		self.semiMajorAxis = [0.0, 0.0, 0.0] # [value, error minus, error plus]
		self.eccentricity = [0.0, 0.0, 0.0] # [value, error minus, error plus]
		self.inclination = [0.0, 0.0, 0.0] # [value, error minus, error plus]
		self.period = [0.0, 0.0, 0.0] # [value, error minus, error plus]
		self.temperature = [0.0, 0.0, 0.0] # [value, error minus, error plus]
		self.discoveryMethod = ""
		self.description = ""
		
		data = inData.split("\n")
		for line in data:
			if line:
				endOpen = 1
				while line[endOpen] != ">": endOpen += 1
				startClose = endOpen+1
				while line[startClose] != "<": startClose += 1
				
				# <tagOpen>tagValue</foo>
				tagOpen = line[1:endOpen]
				tagValue = line[endOpen+1:startClose]
				spaceSplit = tagOpen.split(" ")
				
				# If there are no spaces then there are no error attributes
				if (len(spaceSplit) == 1):
					if tagOpen == "name": self.name.append(tagValue)
					elif tagOpen == "radius": self.radius[0] = float(tagValue)
					elif tagOpen == "mass": self.mass[0] = float(tagValue)
					elif tagOpen == "semimajoraxis": self.semiMajorAxis[0] = float(tagValue)
					elif tagOpen == "eccentricity": self.eccentricity[0] = float(tagValue)
					elif tagOpen == "inclination": self.inclination[0] = float(tagValue)
					elif tagOpen == "period": self.period[0] = float(tagValue)
					elif tagOpen == "temperature": self.temperature[0] = float(tagValue)
					elif tagOpen == "discoverymethod": self.discoveryMethod = tagValue
					elif tagOpen == "discoveryyear": continue # Ignore for now...
					elif tagOpen == "lastupdate": continue # Ignore for now...
					elif tagOpen == "description": self.description = tagValue
					elif tagOpen == "rossitermclaughlin": continue # Ignore for now...
					elif tagOpen == "transittime": continue # Ignore for now...
					elif tagOpen == "longitude": continue # Ignore for now...
					elif tagOpen == "periastron": continue # Ignore for now...
					elif tagOpen == "ascendingnode": continue # Ignore for now...
					else: print "Not implemented in Planet: "+line
				
				# If there are spaces then there are error attributes
				else:
					if spaceSplit[0] == "radius":
						self.radius[0] = float(tagValue)
						self.radius[1] = float(spaceSplit[1][12:-1]) # error minus
						self.radius[2] = float(spaceSplit[2][11:-1]) # error plus
						if spaceSplit[1][:10] != "errorminus" or spaceSplit[2][:9] != "errorplus": print "...something went wrong reading errors in Planet"
					elif spaceSplit[0] == "mass":
						self.mass[0] = float(tagValue)
						self.mass[1] = float(spaceSplit[1][12:-1]) # error minus
						self.mass[2] = float(spaceSplit[2][11:-1]) # error plus
						if spaceSplit[1][:10] != "errorminus" or spaceSplit[2][:9] != "errorplus": print "...something went wrong reading errors in Planet"
					elif spaceSplit[0] == "semimajoraxis":
						self.semiMajorAxis[0] = float(tagValue)
						self.semiMajorAxis[1] = float(spaceSplit[1][12:-1]) # error minus
						self.semiMajorAxis[2] = float(spaceSplit[2][11:-1]) # error plus
						if spaceSplit[1][:10] != "errorminus" or spaceSplit[2][:9] != "errorplus": print "...something went wrong reading errors in Planet"
					elif spaceSplit[0] == "eccentricity":
						self.eccentricity[0] = float(tagValue)
						self.eccentricity[1] = float(spaceSplit[1][12:-1]) # error minus
						self.eccentricity[2] = float(spaceSplit[2][11:-1]) # error plus
						if spaceSplit[1][:10] != "errorminus" or spaceSplit[2][:9] != "errorplus": print "...something went wrong reading errors in Planet"
					elif spaceSplit[0] == "inclination":
						self.inclination[0] = float(tagValue)
						self.inclination[1] = float(spaceSplit[1][12:-1]) # error minus
						self.inclination[2] = float(spaceSplit[2][11:-1]) # error plus
						if spaceSplit[1][:10] != "errorminus" or spaceSplit[2][:9] != "errorplus": print "...something went wrong reading errors in Planet"
					elif spaceSplit[0] == "period":
						self.period[0] = float(tagValue)
						self.period[1] = float(spaceSplit[1][12:-1]) # error minus
						self.period[2] = float(spaceSplit[2][11:-1]) # error plus
						if spaceSplit[1][:10] != "errorminus" or spaceSplit[2][:9] != "errorplus": print "...something went wrong reading errors in Planet"
					elif spaceSplit[0] == "temperature":
						self.temperature[0] = float(tagValue)
						self.temperature[1] = float(spaceSplit[1][12:-1]) # error minus
						self.temperature[2] = float(spaceSplit[2][11:-1]) # error plus
						if spaceSplit[1][:10] != "errorminus" or spaceSplit[2][:9] != "errorplus": print "...something went wrong reading errors in Planet"
					elif spaceSplit[0] == "rossitermclaughlin": continue # Ignore for now...
					elif spaceSplit[0] == "transittime": continue # Ignore for now...
					elif spaceSplit[0] == "longitude": continue # Ignore for now...
					elif spaceSplit[0] == "periastron": continue # Ignore for now...
					elif spaceSplit[0] == "ascendingnode": continue # Ignore for now...
					else: print "Not implemented in Planet: "+line

# -------------------------------------------------------------------------------------
# Function to read through XML files and create appropriate hierarchy object instances
# -------------------------------------------------------------------------------------
def readinXML(inFile):
	xmlDict = {"system":0, "binary":1, "star":2, "planet":3}
	xmlDepth = -1
	xmlObjectList = []
	xmlDataList = []
	parentIds = [0, 0, 0, 0]
	
	currFile = open(inFile, "r")
	for line in currFile:
		line = line.strip()
		endLoc = 1
		while line[endLoc] != ">": endLoc += 1
		
		# If the substring [1:endLoc] is in xmlDict then it is the open tag of an object
		if line[1:endLoc] in xmlDict:
			xmlObjectList.append(line[1:endLoc])
			xmlDataList.append("")
			xmlDepth += 1
		
		# If the substring [2:endLoc] is in xmlDict then it is the end tag of an object
		elif line[2:endLoc] in xmlDict:
			currObject = xmlObjectList.pop(xmlDepth)
			currData = xmlDataList.pop(xmlDepth)
			parentIds[xmlDict[currObject]] += 1
			xmlDepth -= 1
			
			# Create the appropriate class and assign its parent (parent type and parent index)
			if currObject == "system": allSystems.append(System(currData))
			elif currObject == "binary":
				allBinaries.append(Binary(currData))
				allBinaries[-1].parent = [xmlDict[xmlObjectList[-1]], parentIds[xmlDict[xmlObjectList[-1]]]]
			elif currObject == "star":
				allStars.append(Star(currData))
				allStars[-1].parent = [xmlDict[xmlObjectList[-1]], parentIds[xmlDict[xmlObjectList[-1]]]]
			elif currObject == "planet":
				allPlanets.append(Planet(currData))
				allPlanets[-1].parent = [xmlDict[xmlObjectList[-1]], parentIds[xmlDict[xmlObjectList[-1]]]]
		
		# Otherwise the tag is not an object
		else: xmlDataList[xmlDepth] += line+"\n"
	currFile.close()

# --------------------------------------------------------------------------------------------
# Read the data from "allSystems" into the object lists assigning data, parents, and children
# --------------------------------------------------------------------------------------------
# Object lists
allSystems = []
allBinaries = []
allStars = []
allPlanets = []

# Add data from XML to each object list including parent information
ticks = time.time()
readinXML("allSystems.xml")

print "Num Systems = "+str(len(allSystems))
print "Num Binaries = "+str(len(allBinaries))
print "Num Stars = "+str(len(allStars))
print "Num Planets = "+str(len(allPlanets))

# Add child information
for i, binary in enumerate(allBinaries):
	if binary.parent[0] == 0: allSystems[binary.parent[1]].children.append([1, i])
	elif binary.parent[0] == 1: allBinaries[binary.parent[1]].children.append([1, i])
	else: print "...something went wrong assigning children from Binary parent"
for i, star in enumerate(allStars):
	if star.parent[0] == 0: allSystems[star.parent[1]].children.append([2, i])
	elif star.parent[0] == 1: allBinaries[star.parent[1]].children.append([2, i])
	else: print "...something went wrong assigning children from Star parent"
for i, planet in enumerate(allPlanets):
	if planet.parent[0] == 0: allSystems[planet.parent[1]].children.append([3, i])
	elif planet.parent[0] == 1: allBinaries[planet.parent[1]].children.append([3, i])
	elif planet.parent[0] == 2: allStars[planet.parent[1]].children.append([3, i])
	else: print "...something went wrong assigning children from Planet parent"
print "Time to prepare object lists from XML and assign children = "+str(time.time()-ticks)

# ----------------------------------------------------------------------
# Calculate additional data for all class instances and habitable zones
# ----------------------------------------------------------------------
twoPI = math.pi*2.0
earthMass = 0.0031457007
earthRadius = 0.091130294
earthYear = 365.2422

ticks = time.time()
for iBinary in allBinaries:
	# Semi-minor axis based on semi-major axis
	iBinary.semiMinorAxis = iBinary.semiMajorAxis[0]*math.sqrt(1.0-iBinary.eccentricity[0]*iBinary.eccentricity[0])
	
	# Invert all periods
	if iBinary.period[0]: iBinary.period[0] = 1.0/iBinary.period[0]

for iStar in allStars:
	# Radius based on mass conversion from: http://www2.astro.psu.edu/users/rbc/a534/lec18.pdf
	if not iStar.radius[0] and iStar.mass[0]:
		if iStar.mass[0] < 1.0: iStar.radius[0] = iStar.mass[0]**0.57
		else: iStar.radius[0] = iStar.mass[0]**0.8
	
	# Color based on spectral type, hexadecimal values from: http://en.wikipedia.org/wiki/Stellar_classification#Harvard_spectral_classification
	spectralClass = ""
	if iStar.spectralType[0]:
		spectralClass = iStar.spectralType[0][0]
		if spectralClass == "O": iStar.spectralType[1] = "9db4ff" # Blue
		elif spectralClass == "B": iStar.spectralType[1] = "aabfff" # Deep blue white
		elif spectralClass == "A": iStar.spectralType[1] = "cad8ff" # Blue white
		elif spectralClass == "F": iStar.spectralType[1] = "fbf8ff" # White
		elif spectralClass == "G": iStar.spectralType[1] = "fff4e8" # Yellowish white
		elif spectralClass == "K": iStar.spectralType[1] = "ffddb4" # Pale yellow orange
		elif spectralClass == "M": iStar.spectralType[1] = "ffbd6f" # Light orange red
		elif spectralClass == "L": iStar.spectralType[1] = "f84235" # Scarlet
		elif spectralClass == "T": iStar.spectralType[1] = "ba3059" # Magenta
		elif spectralClass == "Y": iStar.spectralType[1] = "605170" # Dark purple
	
	# -------------------------------------------------------------------------
	# Habitable zone based on distance and magV or general table as a fallback
	# -------------------------------------------------------------------------
	# Reassign rootParent until it is not a Binary instance (it will be a System instance which has distance)
	rootParent = iStar.parent
	while rootParent[0] == 1: rootParent = allBinaries[rootParent[1]].parent
	
	if spectralClass:
		# Star instances with magV and root parent distance use equations from: http://www.planetarybiology.com/calculating_habitable_zone.htm
		if iStar.magV[0] and allSystems[rootParent[1]].distance[0]:
			absoluteMagV = iStar.magV[0]-5.0*math.log(allSystems[rootParent[1]].distance[0]/10.0)
			
			# Convert absoluteMagV to mBolStar
			acceptableBolometricMagnitude = False
			if spectralClass == "B":
				absoluteMagV -= 2.0
				acceptableBolometricMagnitude = True
			elif spectralClass == "A":
				absoluteMagV -= 0.3
				acceptableBolometricMagnitude = True
			elif spectralClass == "F":
				absoluteMagV -= 0.15
				acceptableBolometricMagnitude = True
			elif spectralClass == "G":
				absoluteMagV -= 0.4
				acceptableBolometricMagnitude = True
			elif spectralClass == "K":
				absoluteMagV -= 0.8
				acceptableBolometricMagnitude = True
			elif spectralClass == "M":
				absoluteMagV -= 2.0
				acceptableBolometricMagnitude = True
			
			# Only Star instances with a lookup bolometric magnitude value are acceptable
			if acceptableBolometricMagnitude:
				starLuminosity = 10**((absoluteMagV-4.72)/-2.5)
				iStar.habitableInner = math.sqrt(starLuminosity/1.1)
				iStar.habitableOuter = math.sqrt(starLuminosity/0.53)
		
		# Without a distance the general table is used from: http://www.evl.uic.edu/aej/526/project2.html
		if not iStar.habitableInner or not iStar.habitableOuter:
			if spectralClass == "A":
				iStar.habitableInner = 8.5
				iStar.habitableOuter = 12.5
			elif spectralClass == "F":
				iStar.habitableInner = 1.5
				iStar.habitableOuter = 2.2
			elif spectralClass == "G":
				iStar.habitableInner = 0.95
				iStar.habitableOuter = 1.4
			elif spectralClass == "K":
				iStar.habitableInner = 0.38
				iStar.habitableOuter = 0.56
			elif spectralClass == "M":
				iStar.habitableInner = 0.08
				iStar.habitableOuter = 0.12

for iPlanet in allPlanets:
	# Radius based on mass "categories" from: http://www.evl.uic.edu/aej/526/project2.html
	if not iPlanet.radius[0] and iPlanet.mass[0]:
		if iPlanet.mass[0] < earthMass*2.0: iPlanet.radius[0] = iPlanet.mass[0]/(earthMass*2.0)*earthRadius*1.25
		elif iPlanet.mass[0] >= earthMass*2.0 and iPlanet.mass[0] < earthMass*5.0: iPlanet.radius[0] = (iPlanet.mass[0]-earthMass*2.0)/(earthMass*3.0)*earthRadius*2.0
		elif iPlanet.mass[0] >= earthMass*5.0 and iPlanet.mass[0] < earthMass*10.0: iPlanet.radius[0] = (iPlanet.mass[0]-earthMass*5.0)/(earthMass*5.0)*earthRadius*3.0
		elif iPlanet.mass[0] >= earthMass*10.0 and iPlanet.mass[0] < earthMass*30.0: iPlanet.radius[0] = (iPlanet.mass[0]-earthMass*10.0)/(earthMass*20.0)*earthRadius*6.0
		elif iPlanet.mass[0] >= earthMass*30.0: iPlanet.radius[0] = (iPlanet.mass[0]-earthMass*30.0)/(earthMass*270.0)*earthRadius*15.0
	
	# Period and semi-major axis based on each other from: http://hyperphysics.phy-astr.gsu.edu/hbase/kepler.html#c2
	if not iPlanet.period[0] and iPlanet.semiMajorAxis[0]: iPlanet.period[0] = math.sqrt(iPlanet.semiMajorAxis[0]**3)*earthYear
	if not iPlanet.semiMajorAxis[0] and iPlanet.period[0]: iPlanet.semiMajorAxis[0] = (iPlanet.period[0]/earthYear)**(2.0/3.0)
	
	# Semi-minor axis based on semi-major axis
	iPlanet.semiMinorAxis = iPlanet.semiMajorAxis[0]*math.sqrt(1.0-iPlanet.eccentricity[0]*iPlanet.eccentricity[0])
	
	# Invert all periods
	if iPlanet.period[0]: iPlanet.period[0] = 1.0/iPlanet.period[0]
print "Time to create all additional data = "+str(time.time()-ticks)

# --------------------------------------------------
# Determine how much geometry is needed for display
# --------------------------------------------------
# These will be assigned the maximum required
numSpheres = 0
numLineSets = 0
numLights = 0

# This is needed to avoid "jumping" while scaling time
numPeriods = 0

for iSystem in allSystems:
	sphereCount = 0
	lineSetCount = 0
	lightCount = 0
	periodCount = 0
	
	for jChild in iSystem.children:
		if jChild[0] == 1:
			jBinary = allBinaries[jChild[1]]
			if jBinary.period[0] and jBinary.semiMajorAxis[0]: periodCount += 1
			for kChild in jBinary.children:
				if kChild[0] == 1:
					kBinary = allBinaries[kChild[1]]
					if kBinary.period[0] and kBinary.semiMajorAxis[0]: periodCount += 1
					for lChild in kBinary.children:
						if lChild[0] == 1: continue # Ignore for now
						elif lChild[0] == 2:
							lStar = allStars[lChild[1]]
							if lStar.radius[0]: sphereCount += 1
							if lStar.habitableInner and lStar.habitableOuter: lineSetCount += 1
							lightCount += 1
							for mChild in lStar.children:
								mPlanet = allPlanets[mChild[1]]
								if mPlanet.period[0] and mPlanet.semiMajorAxis[0]:
									sphereCount += 1
									lineSetCount += 1
									periodCount += 1
						elif lChild[0] == 3:
							lPlanet = allPlanets[lChild[1]]
							if lPlanet.period[0] and lPlanet.semiMajorAxis[0]:
								sphereCount += 1
								lineSetCount += 1
								periodCount += 1
				elif kChild[0] == 2:
					kStar = allStars[kChild[1]]
					if kStar.radius[0]: sphereCount += 1
					if kStar.habitableInner and kStar.habitableOuter: lineSetCount += 1
					lightCount += 1
					for lChild in kStar.children:
						lPlanet = allPlanets[lChild[1]]
						if lPlanet.period[0] and lPlanet.semiMajorAxis[0]:
							sphereCount += 1
							lineSetCount += 1
							periodCount += 1
				elif kChild[0] == 3:
					kPlanet = allPlanets[kChild[1]]
					if kPlanet.period[0] and kPlanet.semiMajorAxis[0]:
						sphereCount += 1
						lineSetCount += 1
						periodCount += 1
		elif jChild[0] == 2:
			jStar = allStars[jChild[1]]
			if jStar.radius[0]: sphereCount += 1
			if jStar.habitableInner and jStar.habitableOuter: lineSetCount += 1
			lightCount += 1
			for kChild in jStar.children:
				kPlanet = allPlanets[kChild[1]]
				if kPlanet.period[0] and kPlanet.semiMajorAxis[0]:
					sphereCount += 1
					lineSetCount += 1
					periodCount += 1
		elif jChild[0] == 3:
			jPlanet = allPlanets[jChild[1]]
			if jPlanet.period[0] and jPlanet.semiMajorAxis[0]:
				sphereCount += 1
				# Orphan planets do not have a lineSet
	
	numSpheres = sphereCount if sphereCount > numSpheres else numSpheres
	numLineSets = lineSetCount if lineSetCount > numLineSets else numLineSets
	numLights = lightCount if lightCount > numLights else numLights
	numPeriods = periodCount if periodCount > numPeriods else numPeriods

# This is needed to avoid "jumping" while scaling time
periodTheta = []
for i in range(numPeriods): periodTheta.append(0.0)

print "Geometry: numSpheres = "+str(numSpheres)+", numLineSets = "+str(numLineSets)+", numLights = "+str(numLights)

# ------------------------------------------
# Create the appropriate amount of geometry
# ------------------------------------------
ticks = time.time()

# Both Star and Planet instances use this
allSpheres = []
for i in range(numSpheres):
	allSpheres.append(SphereShape.create(1.0, 4))
	allSpheres[i].setVisible(False)

# Both Star and Planet instances use this
allLineSets = []
allLineSetSegments = []
allLineSetCaps = []
segmentsPerLineSet = 64
thetaInc = twoPI/segmentsPerLineSet
for i in range(numLineSets):
	allLineSets.append(LineSet.create())
	allLineSetSegments.append([])
	allLineSetCaps.append([])
	for j in range(segmentsPerLineSet):
		allLineSetSegments[i].append(allLineSets[-1].addLine())
		allLineSetCaps[i].append(SphereShape.create(1.0, 2))
		allLineSetCaps[i][j].setVisible(False)
	allLineSets[i].setVisible(False)

# Star instances use this
allLights = []
for i in range(numLights):
	allLights.append(Light.create())
	allLights[i].setColor(Color("#00000000")) # Hack to disable lights
	allLights[i].setEnabled(True)

print "Time to create all geometry = "+str(time.time()-ticks)

# ----------------------------------------------------
# System instance control and normalization variables
# ----------------------------------------------------
# Arbitrarily set as 0 (will be overwritten)
currSystem = 0

# Normalizing variables
maxPeriod = 0.0 # The Binary or Planet instance that maximizes this value (divided by twoPI) will orbit once per second
maxBinaryOrbit = 0.0 # The Binary instance that maximizes this value will have a semi-minor axis of 1.0
maxStarRadius = 0.0 # The Star instance that maximizes this value will have a radius of 1.0
maxPlanetRadius = 0.0 # The Planet instance that maximizes this value will have a radius of 1.0
maxPlanetOrbit = 0.0 # The Planet instance that maximizes this value will have a semi-minor axis of 1.0

# User control variables
userPeriod = 1.0
userBinaryOrbit = 10.0 # Normalized 1.0 is too small
userStarRadius = 1.0
userPlanetRadius = 1.0
userPlanetOrbit = 10.0 # Normalized 1.0 is too small
constantDistance = False # Enabling this removes control between Binary and Planet distances

# --------------------------------------------------------------------
# Create a dictionary that gives System instance indices given a name
# --------------------------------------------------------------------
systemDict = {}
for i, iSystem in enumerate(allSystems):
	for jName in iSystem.name:
		if jName not in systemDict: systemDict[jName] = i
		else: print "...duplicate name: "+jName

# -------------------------------------------
# Function to scale System instance geometry
# -------------------------------------------
def scaleSystemGeometry():
	global maxBinaryOrbit, maxPlanetOrbit
	
	if constantDistance:
		maxBinaryOrbit = max(maxBinaryOrbit, maxPlanetOrbit)
		maxPlanetOrbit = max(maxBinaryOrbit, maxPlanetOrbit)
	
	sphereCount = 0
	lineSetCount = 0
	lightCount = 0
	
	for iChild in allSystems[currSystem].children:
		if iChild[0] == 1:
			iBinary = allBinaries[iChild[1]]
			for jChild in iBinary.children:
				if jChild[0] == 1:
					jBinary = allBinaries[jChild[1]]
					for kChild in jBinary.children:
						if kChild[0] == 1: continue # Ignore for now...
						elif kChild[0] == 2:
							kStar = allStars[kChild[1]]
							if kStar.radius[0]:
								linearScale = kStar.radius[0]/maxStarRadius*userStarRadius
								# Sphere positioned in onUpdate()
								allSpheres[sphereCount].setScale(Vector3(linearScale, linearScale, linearScale))
								sphereCount += 1
							
							if kStar.habitableInner and kStar.habitableOuter and maxPlanetOrbit:
								hRadius = (kStar.habitableOuter+kStar.habitableInner)*0.5
								linearScale = hRadius/maxPlanetOrbit*userPlanetOrbit
				
								hThickness = (kStar.habitableOuter-kStar.habitableInner)/maxPlanetOrbit*userPlanetOrbit
								halfThickness = hThickness*0.5
								
								# Lineset positioned in onUpdate()
								allLineSets[lineSetCount].setScale(Vector3(1.0, 1.0, hThickness*0.0000001))
								theta = 0.0
								previousPoint = Vector3(linearScale, 0.0, 0.0)
								for segs in range(segmentsPerLineSet):
									theta += thetaInc
									currPoint = Vector3(cos(theta)*linearScale, sin(theta)*linearScale, 0.0)
									allLineSetSegments[lineSetCount][segs].setStart(previousPoint)
									allLineSetSegments[lineSetCount][segs].setEnd(currPoint)
									allLineSetSegments[lineSetCount][segs].setThickness(hThickness)
									# Sphere positioned in onUpdate()
									allLineSetCaps[lineSetCount][segs].setScale(Vector3(halfThickness, halfThickness, halfThickness*0.0000001))
									previousPoint = currPoint
								lineSetCount += 1
							
							# Light positioned in onUpdate()
							lightCount += 1
							
							for lChild in kStar.children:
								lPlanet = allPlanets[lChild[1]]
								if lPlanet.semiMajorAxis[0] and lPlanet.period[0]:
									linearScale = lPlanet.radius[0]/maxPlanetRadius*userPlanetRadius
									quarterLinearScale = linearScale*0.25
									eighthLinearScale = linearScale*0.125
									
									# Sphere positioned in onUpdate()
									allSpheres[sphereCount].setScale(Vector3(linearScale, linearScale, linearScale))
									sphereCount += 1
									
									# Lineset positioned in onUpdate()
									theta = 0.0
									previousPoint = Vector3(lPlanet.semiMajorAxis[0]/maxPlanetOrbit*userPlanetOrbit, 0.0, 0.0)
									for segs in range(segmentsPerLineSet):
										theta += thetaInc
										currPoint = Vector3(cos(theta)*lPlanet.semiMajorAxis[0]/maxPlanetOrbit*userPlanetOrbit, sin(theta)*lPlanet.semiMinorAxis/maxPlanetOrbit*userPlanetOrbit, 0.0)
										allLineSetSegments[lineSetCount][segs].setStart(previousPoint)
										allLineSetSegments[lineSetCount][segs].setEnd(currPoint)
										allLineSetSegments[lineSetCount][segs].setThickness(quarterLinearScale)
										# Sphere positioned in onUpdate()
										allLineSetCaps[lineSetCount][segs].setScale(Vector3(eighthLinearScale, eighthLinearScale, eighthLinearScale))
										previousPoint = currPoint
									lineSetCount += 1
						
						elif kChild[0] == 3:
							kPlanet = allPlanets[kChild[1]]
							if kPlanet.period[0] and kPlanet.semiMajorAxis[0]:
								linearScale = kPlanet.radius[0]/maxPlanetRadius*userPlanetRadius
								quarterLinearScale = linearScale*0.25
								eighthLinearScale = linearScale*0.125
								
								# Sphere positioned in onUpdate()
								allSpheres[sphereCount].setScale(Vector3(linearScale, linearScale, linearScale))
								sphereCount += 1
								
								# Lineset positioned in onUpdate()
								theta = 0.0
								previousPoint = Vector3(kPlanet.semiMajorAxis[0]/maxPlanetOrbit*userPlanetOrbit, 0.0, 0.0)
								for segs in range(segmentsPerLineSet):
									theta += thetaInc
									currPoint = Vector3(cos(theta)*kPlanet.semiMajorAxis[0]/maxPlanetOrbit*userPlanetOrbit, sin(theta)*kPlanet.semiMinorAxis/maxPlanetOrbit*userPlanetOrbit, 0.0)
									allLineSetSegments[lineSetCount][segs].setStart(previousPoint)
									allLineSetSegments[lineSetCount][segs].setEnd(currPoint)
									allLineSetSegments[lineSetCount][segs].setThickness(quarterLinearScale)
									# Sphere positioned in onUpdate()
									allLineSetCaps[lineSetCount][segs].setScale(Vector3(eighthLinearScale, eighthLinearScale, eighthLinearScale))
									previousPoint = currPoint
								lineSetCount += 1
				
				elif jChild[0] == 2:
					jStar = allStars[jChild[1]]
					if jStar.radius[0]:
						linearScale = jStar.radius[0]/maxStarRadius*userStarRadius
						# Sphere positioned in onUpdate()
						allSpheres[sphereCount].setScale(Vector3(linearScale, linearScale, linearScale))
						sphereCount += 1
						
					if jStar.habitableInner and jStar.habitableOuter and maxPlanetOrbit:
						hRadius = (jStar.habitableOuter+jStar.habitableInner)*0.5
						linearScale = hRadius/maxPlanetOrbit*userPlanetOrbit
				
						hThickness = (jStar.habitableOuter-jStar.habitableInner)/maxPlanetOrbit*userPlanetOrbit
						halfThickness = hThickness*0.5
						
						# Lineset positioned in onUpdate()
						allLineSets[lineSetCount].setScale(Vector3(1.0, 1.0, hThickness*0.0000001))
						theta = 0.0
						previousPoint = Vector3(linearScale, 0.0, 0.0)
						for segs in range(segmentsPerLineSet):
							theta += thetaInc
							currPoint = Vector3(cos(theta)*linearScale, sin(theta)*linearScale, 0.0)
							allLineSetSegments[lineSetCount][segs].setStart(previousPoint)
							allLineSetSegments[lineSetCount][segs].setEnd(currPoint)
							allLineSetSegments[lineSetCount][segs].setThickness(hThickness)
							# Sphere positioned in onUpdate()
							allLineSetCaps[lineSetCount][segs].setScale(Vector3(halfThickness, halfThickness, halfThickness*0.0000001))
							previousPoint = currPoint
						lineSetCount += 1
					
					# Light positioned in onUpdate()
					lightCount += 1
					
					for kChild in jStar.children:
						kPlanet = allPlanets[kChild[1]]
						if kPlanet.semiMajorAxis[0] and kPlanet.period[0]:
							linearScale = kPlanet.radius[0]/maxPlanetRadius*userPlanetRadius
							quarterLinearScale = linearScale*0.25
							eighthLinearScale = linearScale*0.125
							
							# Sphere positioned in onUpdate()
							allSpheres[sphereCount].setScale(Vector3(linearScale, linearScale, linearScale))
							sphereCount += 1
							
							# Lineset positioned in onUpdate()
							theta = 0.0
							previousPoint = Vector3(kPlanet.semiMajorAxis[0]/maxPlanetOrbit*userPlanetOrbit, 0.0, 0.0)
							for segs in range(segmentsPerLineSet):
								theta += thetaInc
								currPoint = Vector3(cos(theta)*kPlanet.semiMajorAxis[0]/maxPlanetOrbit*userPlanetOrbit, sin(theta)*kPlanet.semiMinorAxis/maxPlanetOrbit*userPlanetOrbit, 0.0)
								allLineSetSegments[lineSetCount][segs].setStart(previousPoint)
								allLineSetSegments[lineSetCount][segs].setEnd(currPoint)
								allLineSetSegments[lineSetCount][segs].setThickness(quarterLinearScale)
								# Sphere positioned in onUpdate()
								allLineSetCaps[lineSetCount][segs].setScale(Vector3(eighthLinearScale, eighthLinearScale, eighthLinearScale))
								previousPoint = currPoint
							lineSetCount += 1
				
				elif jChild[0] == 3:
					jPlanet = allPlanets[jChild[1]]
					if jPlanet.period[0] and jPlanet.semiMajorAxis[0]:
						linearScale = jPlanet.radius[0]/maxPlanetRadius*userPlanetRadius
						quarterLinearScale = linearScale*0.25
						eighthLinearScale = linearScale*0.125
						
						# Sphere positioned in onUpdate()
						allSpheres[sphereCount].setScale(Vector3(linearScale, linearScale, linearScale))
						sphereCount += 1
						
						allLineSets[lineSetCount].setPosition(Vector3(0.0, 0.0, 0.0)) # No need to position this lineset in onUpdate()
						theta = 0.0
						previousPoint = Vector3(jPlanet.semiMajorAxis[0]/maxPlanetOrbit*userPlanetOrbit, 0.0, 0.0)
						for segs in range(segmentsPerLineSet):
							theta += thetaInc
							currPoint = Vector3(cos(theta)*jPlanet.semiMajorAxis[0]/maxPlanetOrbit*userPlanetOrbit, sin(theta)*jPlanet.semiMinorAxis/maxPlanetOrbit*userPlanetOrbit, 0.0)
							allLineSetSegments[lineSetCount][segs].setStart(previousPoint)
							allLineSetSegments[lineSetCount][segs].setEnd(currPoint)
							allLineSetSegments[lineSetCount][segs].setThickness(quarterLinearScale)
							allLineSetCaps[lineSetCount][segs].setPosition(Vector3(0.0, 0.0, 0.0)+previousPoint) # No need to position this sphere in onUpdate()
							allLineSetCaps[lineSetCount][segs].setScale(Vector3(eighthLinearScale, eighthLinearScale, eighthLinearScale))
							previousPoint = currPoint
						lineSetCount += 1
		
		elif iChild[0] == 2:
			iStar = allStars[iChild[1]]
			if iStar.radius[0]:
				linearScale = iStar.radius[0]/maxStarRadius*userStarRadius
				allSpheres[sphereCount].setPosition(Vector3(0.0, 0.0, 0.0)) # No need to position this sphere in onUpdate()
				allSpheres[sphereCount].setScale(Vector3(linearScale, linearScale, linearScale))
				sphereCount += 1
			
			if iStar.habitableInner and iStar.habitableOuter and maxPlanetOrbit:
				hRadius = (iStar.habitableOuter+iStar.habitableInner)*0.5
				linearScale = hRadius/maxPlanetOrbit*userPlanetOrbit
				
				hThickness = (iStar.habitableOuter-iStar.habitableInner)/maxPlanetOrbit*userPlanetOrbit
				halfThickness = hThickness*0.5
				
				allLineSets[lineSetCount].setPosition(Vector3(0.0, 0.0, 0.0)) # No need to position this lineset in onUpdate()
				allLineSets[lineSetCount].setScale(Vector3(1.0, 1.0, hThickness*0.0000001))
				theta = 0.0
				previousPoint = Vector3(linearScale, 0.0, 0.0)
				for segs in range(segmentsPerLineSet):
					theta += thetaInc
					currPoint = Vector3(cos(theta)*linearScale, sin(theta)*linearScale, 0.0)
					allLineSetSegments[lineSetCount][segs].setStart(previousPoint)
					allLineSetSegments[lineSetCount][segs].setEnd(currPoint)
					allLineSetSegments[lineSetCount][segs].setThickness(hThickness)
					allLineSetCaps[lineSetCount][segs].setPosition(previousPoint) # No need to position this sphere in onUpdate()
					allLineSetCaps[lineSetCount][segs].setScale(Vector3(halfThickness, halfThickness, halfThickness*0.0000001))
					previousPoint = currPoint
				lineSetCount += 1
			
			allLights[lightCount].setPosition(Vector3(0.0, 0.0, 0.0)) # No need to position this light in onUpdate()
			lightCount += 1
			
			for jChild in iStar.children:
				jPlanet = allPlanets[jChild[1]]
				if jPlanet.semiMajorAxis[0] and jPlanet.period[0]:
					linearScale = jPlanet.radius[0]/maxPlanetRadius*userPlanetRadius
					quarterLinearScale = linearScale*0.25
					eighthLinearScale = linearScale*0.125
					
					allSpheres[sphereCount].setPosition(Vector3(0.0, 0.0, 0.0)) # No need to position this sphere in onUpdate()
					allSpheres[sphereCount].setScale(Vector3(linearScale, linearScale, linearScale))
					sphereCount += 1
					
					allLineSets[lineSetCount].setPosition(Vector3(0.0, 0.0, 0.0)) # No need to position this lineset in onUpdate()
					theta = 0.0
					previousPoint = Vector3(jPlanet.semiMajorAxis[0]/maxPlanetOrbit*userPlanetOrbit, 0.0, 0.0)
					for segs in range(segmentsPerLineSet):
						theta += thetaInc
						currPoint = Vector3(cos(theta)*jPlanet.semiMajorAxis[0]/maxPlanetOrbit*userPlanetOrbit, sin(theta)*jPlanet.semiMinorAxis/maxPlanetOrbit*userPlanetOrbit, 0.0)
						allLineSetSegments[lineSetCount][segs].setStart(previousPoint)
						allLineSetSegments[lineSetCount][segs].setEnd(currPoint)
						allLineSetSegments[lineSetCount][segs].setThickness(quarterLinearScale)
						allLineSetCaps[lineSetCount][segs].setPosition(previousPoint) # No need to position this sphere in onUpdate()
						allLineSetCaps[lineSetCount][segs].setScale(Vector3(eighthLinearScale, eighthLinearScale, eighthLinearScale))
						previousPoint = currPoint
					lineSetCount += 1
		
		elif iChild[0] == 3:
			iPlanet = allPlanets[iChild[1]]
			if iPlanet.period[0] and iPlanet.semiMajorAxis[0]:
				linearScale = iPlanet.radius[0]/maxPlanetRadius
				allSpheres[sphereCount].setPosition(Vector3(0.0, 0.0, 0.0)) # No need to position this sphere in onUpdate()
				allSpheres[sphereCount].setScale(Vector3(linearScale, linearScale, linearScale))
				sphereCount += 1

# -----------------------------------------------
# Object Description and Missing Data Containers
# -----------------------------------------------
objectDescriptions = wf.createContainer("ObjectDescriptions", uiRoot, ContainerLayout.LayoutFree)
objectDescriptions.setStyle("fill:#20202080")
objectDescriptions.setSize(Vector2(5464, 768))
objectDescriptions.setPosition(Vector2(8196, 0))
if clippingOn: objectDescriptions.setClippingEnabled(True) # So far not needed
objectDescriptions.setAutosize(False)

objectDescriptionLabels = []
for i in range(10):
	objectDescriptionLabels.append(wf.createLabel("ObjectDescriptionLabelsLabel:"+str(i), objectDescriptions, " ")) # Apparently empty string is bad
	objectDescriptionLabels[i].setStyle("color:#ff8080; font:fonts/arial.ttf 24")
	objectDescriptionLabels[i].setPosition(Vector2(8, 12+i*32))
objectDescriptionLabels[0].setText("Planet Descriptions:")

missingData = wf.createContainer("MissingData", uiRoot, ContainerLayout.LayoutFree)
missingData.setStyle("fill:#20202080")
missingData.setSize(Vector2(2732, 768))
missingData.setPosition(Vector2(13660, 0))
if clippingOn: missingData.setClippingEnabled(True) # So far not needed
missingData.setAutosize(False)

missingDataLabels = []
for i in range(maxMultChildren):
	missingDataLabels.append(wf.createLabel("MissingDataLabelsLabel:"+str(i), missingData, " ")) # Apparently empty string is bad
	missingDataLabels[i].setStyle("color:#ff8080; font:fonts/arial.ttf 24")
	missingDataLabels[i].setPosition(Vector2(8, 12+i*32))
missingDataLabels[0].setText("Missing Data:")

# ----------------------------------------
# Function to set current System instance
# ----------------------------------------
def setSystem(inName):
	global currSystem, maxPeriod, maxBinaryOrbit, maxStarRadius, maxPlanetRadius, maxPlanetOrbit
	
	# Reset maximums
	maxPeriod = 0.0
	maxBinaryOrbit = 0.0
	maxStarRadius = 0.0
	maxPlanetRadius = 0.0
	maxPlanetOrbit = 0.0
	
	# Assign currSystem as the selected system
	currSystem = systemDict[inName]
	
	# Hide all geometry
	for i in range(numSpheres): allSpheres[i].setVisible(False)
	for i in range(numLineSets):
		allLineSets[i].setVisible(False)
		for j in range(segmentsPerLineSet): allLineSetCaps[i][j].setVisible(False)
	for i in range(numLights): allLights[i].setColor(Color("#00000000")) # Hack to disable lights
	
	sphereCount = 0
	lineSetCount = 0
	lightCount = 0
	
	# Make current system geometry visible and assign maximums
	for iChild in allSystems[currSystem].children:
		if iChild[0] == 1:
			iBinary = allBinaries[iChild[1]]
			if iBinary.period[0] and iBinary.semiMajorAxis[0]: maxPeriod = iBinary.period[0] if iBinary.period[0] > maxPeriod else maxPeriod
			maxBinaryOrbit = iBinary.semiMinorAxis if iBinary.semiMinorAxis > maxBinaryOrbit else maxBinaryOrbit
			
			for jChild in iBinary.children:
				if jChild[0] == 1:
					jBinary = allBinaries[jChild[1]]
					if jBinary.period[0] and jBinary.semiMajorAxis[0]: maxPeriod = jBinary.period[0] if jBinary.period[0] > maxPeriod else maxPeriod
					maxBinaryOrbit = jBinary.semiMinorAxis if jBinary.semiMinorAxis > maxBinaryOrbit else maxBinaryOrbit
					
					for kChild in jBinary.children:
						if kChild[0] == 1: continue # Ignore for now...
						elif kChild[0] == 2:
							kStar = allStars[kChild[1]]
							if kStar.radius[0]:
								if kStar.spectralType[1]: allSpheres[sphereCount].setEffect("colored -e #"+kStar.spectralType[1]+"80 -t")
								else: allSpheres[sphereCount].setEffect("colored -e #00800080 -t") # Invalid / missing spectral type
								allSpheres[sphereCount].setVisible(True)
								sphereCount += 1
								
								maxStarRadius = kStar.radius[0] if kStar.radius[0] > maxStarRadius else maxStarRadius
							
							if kStar.habitableInner and kStar.habitableOuter:
								allLineSets[lineSetCount].setEffect("colored -e #00ff80")
								allLineSets[lineSetCount].setVisible(True)
								for segs in range(segmentsPerLineSet):
									allLineSetCaps[lineSetCount][segs].setEffect("colored -e #00ff80")
									allLineSetCaps[lineSetCount][segs].setVisible(True)
								lineSetCount += 1
							
							if kStar.spectralType[1]: allLights[lightCount].setColor(Color("#"+kStar.spectralType[1]))
							else: allLights[lightCount].setColor(Color("#008000")) # Invalid / missing spectral type
							lightCount += 1
							
							for lChild in kStar.children:
								lPlanet = allPlanets[lChild[1]]
								if lPlanet.period[0] and lPlanet.semiMajorAxis[0]:
									allSpheres[sphereCount].setEffect("colored -d #ffffff")
									allSpheres[sphereCount].setVisible(True)
									sphereCount += 1
									
									allLineSets[lineSetCount].setEffect("colored -e #ffffff")
									allLineSets[lineSetCount].setVisible(True)
									for segs in range(segmentsPerLineSet):
										allLineSetCaps[lineSetCount][segs].setEffect("colored -e #ffffff")
										allLineSetCaps[lineSetCount][segs].setVisible(True)
									lineSetCount += 1
									
									maxPeriod = lPlanet.period[0] if lPlanet.period[0] > maxPeriod else maxPeriod
									maxPlanetRadius = lPlanet.radius[0] if lPlanet.radius[0] > maxPlanetRadius else maxPlanetRadius
									maxPlanetOrbit = lPlanet.semiMinorAxis if lPlanet.semiMinorAxis > maxPlanetOrbit else maxPlanetOrbit
						
						elif kChild[0] == 3:
							kPlanet = allPlanets[kChild[1]]
							if kPlanet.period[0] and kPlanet.semiMajorAxis[0]:
								allSpheres[sphereCount].setEffect("colored -d #ffffff")
								allSpheres[sphereCount].setVisible(True)
								sphereCount += 1
								
								allLineSets[lineSetCount].setEffect("colored -e #ffffff")
								allLineSets[lineSetCount].setVisible(True)
								for segs in range(segmentsPerLineSet):
									allLineSetCaps[lineSetCount][segs].setEffect("colored -e #ffffff")
									allLineSetCaps[lineSetCount][segs].setVisible(True)
								lineSetCount += 1
								
								maxPeriod = kPlanet.period[0] if kPlanet.period[0] > maxPeriod else maxPeriod
								maxPlanetRadius = kPlanet.radius[0] if kPlanet.radius[0] > maxPlanetRadius else maxPlanetRadius
								maxPlanetOrbit = kPlanet.semiMinorAxis if kPlanet.semiMinorAxis > maxPlanetOrbit else maxPlanetOrbit
				
				elif jChild[0] == 2:
					jStar = allStars[jChild[1]]
					if jStar.radius[0]:
						if jStar.spectralType[1]: allSpheres[sphereCount].setEffect("colored -e #"+jStar.spectralType[1]+"80 -t")
						else: allSpheres[sphereCount].setEffect("colored -e #00800080 -t") # Invalid / missing spectral type
						allSpheres[sphereCount].setVisible(True)
						sphereCount += 1
						
						maxStarRadius = jStar.radius[0] if jStar.radius[0] > maxStarRadius else maxStarRadius
					
					if jStar.habitableInner and jStar.habitableOuter:
						allLineSets[lineSetCount].setEffect("colored -e #00ff80")
						allLineSets[lineSetCount].setVisible(True)
						for segs in range(segmentsPerLineSet):
							allLineSetCaps[lineSetCount][segs].setEffect("colored -e #00ff80")
							allLineSetCaps[lineSetCount][segs].setVisible(True)
						lineSetCount += 1
					
					if jStar.spectralType[1]: allLights[lightCount].setColor(Color("#"+jStar.spectralType[1]))
					else: allLights[lightCount].setColor(Color("#008000")) # Invalid / missing spectral type
					lightCount += 1
					
					for kChild in jStar.children:
						kPlanet = allPlanets[kChild[1]]
						if kPlanet.period[0] and kPlanet.semiMajorAxis[0]:
							allSpheres[sphereCount].setEffect("colored -d #ffffff")
							allSpheres[sphereCount].setVisible(True)
							sphereCount += 1
							
							allLineSets[lineSetCount].setEffect("colored -e #ffffff")
							allLineSets[lineSetCount].setVisible(True)
							for segs in range(segmentsPerLineSet):
								allLineSetCaps[lineSetCount][segs].setEffect("colored -e #ffffff")
								allLineSetCaps[lineSetCount][segs].setVisible(True)
							lineSetCount += 1
							
							maxPeriod = kPlanet.period[0] if kPlanet.period[0] > maxPeriod else maxPeriod
							maxPlanetRadius = kPlanet.radius[0] if kPlanet.radius[0] > maxPlanetRadius else maxPlanetRadius
							maxPlanetOrbit = kPlanet.semiMinorAxis if kPlanet.semiMinorAxis > maxPlanetOrbit else maxPlanetOrbit
				
				elif jChild[0] == 3:
					jPlanet = allPlanets[jChild[1]]
					if jPlanet.period[0] and jPlanet.semiMajorAxis[0]:
						allSpheres[sphereCount].setEffect("colored -d #ffffff")
						allSpheres[sphereCount].setVisible(True)
						sphereCount += 1
						
						allLineSets[lineSetCount].setEffect("colored -e #ffffff")
						allLineSets[lineSetCount].setVisible(True)
						for segs in range(segmentsPerLineSet):
							allLineSetCaps[lineSetCount][segs].setEffect("colored -e #ffffff")
							allLineSetCaps[lineSetCount][segs].setVisible(True)
						lineSetCount += 1
						
						maxPeriod = jPlanet.period[0] if jPlanet.period[0] > maxPeriod else maxPeriod
						maxPlanetRadius = jPlanet.radius[0] if jPlanet.radius[0] > maxPlanetRadius else maxPlanetRadius
						maxPlanetOrbit = jPlanet.semiMinorAxis if jPlanet.semiMinorAxis > maxPlanetOrbit else maxPlanetOrbit
		
		elif iChild[0] == 2:
			iStar = allStars[iChild[1]]
			if iStar.radius[0]:
				if iStar.spectralType[1]: allSpheres[sphereCount].setEffect("colored -e #"+iStar.spectralType[1]+"80 -t")
				else: allSpheres[sphereCount].setEffect("colored -e #00800080 -t") # Invalid / missing spectral type
				allSpheres[sphereCount].setVisible(True)
				sphereCount += 1
				
				maxStarRadius = iStar.radius[0] if iStar.radius[0] > maxStarRadius else maxStarRadius
			
			if iStar.habitableInner and iStar.habitableOuter:
				allLineSets[lineSetCount].setEffect("colored -e #00ff80")
				allLineSets[lineSetCount].setVisible(True)
				for segs in range(segmentsPerLineSet):
					allLineSetCaps[lineSetCount][segs].setEffect("colored -e #00ff80")
					allLineSetCaps[lineSetCount][segs].setVisible(True)
				lineSetCount += 1
			
			if iStar.spectralType[1]: allLights[lightCount].setColor(Color("#"+iStar.spectralType[1]))
			else: allLights[lightCount].setColor(Color("#008000")) # Invalid / missing spectral type
			lightCount += 1
			
			for jChild in iStar.children:
				jPlanet = allPlanets[jChild[1]]
				if jPlanet.period[0] and jPlanet.semiMajorAxis[0]:
					allSpheres[sphereCount].setEffect("colored -d #ffffff")
					allSpheres[sphereCount].setVisible(True)
					sphereCount += 1
					
					allLineSets[lineSetCount].setEffect("colored -e #ffffff")
					allLineSets[lineSetCount].setVisible(True)
					for segs in range(segmentsPerLineSet):
						allLineSetCaps[lineSetCount][segs].setEffect("colored -e #ffffff")
						allLineSetCaps[lineSetCount][segs].setVisible(True)
					lineSetCount += 1
					
					maxPeriod = jPlanet.period[0] if jPlanet.period[0] > maxPeriod else maxPeriod
					maxPlanetRadius = jPlanet.radius[0] if jPlanet.radius[0] > maxPlanetRadius else maxPlanetRadius
					maxPlanetOrbit = jPlanet.semiMinorAxis if jPlanet.semiMinorAxis > maxPlanetOrbit else maxPlanetOrbit
		
		elif iChild[0] == 3:
			iPlanet = allPlanets[iChild[1]]
			if iPlanet.period[0] and iPlanet.semiMajorAxis[0]:
				allSpheres[sphereCount].setEffect("colored -d #ffffff")
				allSpheres[sphereCount].setVisible(True)
				sphereCount += 1
				
				maxPlanetRadius = jPlanet.radius[0] if jPlanet.radius[0] > maxPlanetRadius else maxPlanetRadius
	
	# Scale and position geometry
	scaleSystemGeometry()
	
	# Alert missing geometry due to inadequate data and show descriptions if any
	objectDescriptionCount = 1
	missingObjectCount = 1
	for iChild in allSystems[currSystem].children:
		if iChild[0] == 1:
			iBinary = allBinaries[iChild[1]]
			for jChild in iBinary.children:
				if jChild[0] == 1:
					jBinary = allBinaries[jChild[1]]
					for kChild in jBinary.children:
						if kChild[0] == 1: continue # Ignore for now
						elif kChild[0] == 2:
							kStar = allStars[kChild[1]]
							if not kStar.radius[0]:
								starString = "Star: "+kStar.name[0]
								if kStar.spectralType[0]: starString += "; Spectral Type: "+kStar.spectralType[0]
								else: starString += "; Unknown Spectral Type"
								starString += "; No Radius / Mass data"
								missingDataLabels[missingObjectCount].setText(tab+starString)
								missingDataLabels[missingObjectCount].setVisible(True)
								missingObjectCount += 1
								print "Missing "+starString
							
							if not kStar.habitableInner or not kStar.habitableOuter or not maxPlanetOrbit:
								missingDataLabels[missingObjectCount].setText(tab+tab+kStar.name[0]+"'s habitiable zone, insufficient data for Luminosity and look-up table methods")
								missingDataLabels[missingObjectCount].setVisible(True)
								missingObjectCount += 1
								print "Missing Star: "+kStar.name[0]+"'s habitable zone"
							
							for lChild in kStar.children:
								lPlanet = allPlanets[lChild[1]]
								if not lPlanet.period[0] or not lPlanet.semiMajorAxis[0]:
									planetString = "Planet: "+lPlanet.name[0]
									if lPlanet.radius[0]:
										planetString += "; Radius: "+str(lPlanet.radius[0])
										if lPlanet.radius[1] or lPlanet.radius[2]: planetString += " (error -"+str(lPlanet.radius[1])+" +"+str(lPlanet.radius[2])+")"
									else: planetString += "; No Radius / Mass data"
									if lPlanet.semiMajorAxis[0]:
										planetString += "; Semi-major Axis: "+str(lPlanet.semiMajorAxis[0])
										if lPlanet.semiMajorAxis[1] or lPlanet.semiMajorAxis[2]: planetString += " (error -"+str(lPlanet.semiMajorAxis[1])+" +"+str(lPlanet.semiMajorAxis[2])+")"
									else: planetString += "; No Semi-major Axis data"
									if lPlanet.period[0]:
										planetString += "; 1.0/Period: "+str(lPlanet.period[0])
										if lPlanet.period[1] or lPlanet.period[2]: planetString += " (error -"+str(lPlanet.period[1])+" +"+str(lPlanet.period[2])+")"
									else: planetString += "; No Period data"
									missingDataLabels[missingObjectCount].setText(tab+tab+planetString)
									missingDataLabels[missingObjectCount].setVisible(True)
									missingObjectCount += 1
									print "Missing "+planetString
								
								if lPlanet.description:
									objectDescriptionLabels[objectDescriptionCount].setText(tab+lPlanet.description)
									objectDescriptionLabels[objectDescriptionCount].setVisible(True)
									objectDescriptionCount += 1
						
						elif kChild[0] == 3:
							kPlanet = allPlanets[kChild[1]]
							if not kPlanet.period[0] or not kPlanet.semiMajorAxis[0]:
								planetString = "Planet: "+kPlanet.name[0]
								if kPlanet.radius[0]:
									planetString += "; Radius: "+str(kPlanet.radius[0])
									if kPlanet.radius[1] or kPlanet.radius[2]: planetString += " (error -"+str(kPlanet.radius[1])+" +"+str(kPlanet.radius[2])+")"
								else: planetString += "; No Radius / Mass data"
								if kPlanet.semiMajorAxis[0]:
									planetString += "; Semi-major Axis: "+str(kPlanet.semiMajorAxis[0])
									if kPlanet.semiMajorAxis[1] or kPlanet.semiMajorAxis[2]: planetString += " (error -"+str(kPlanet.semiMajorAxis[1])+" +"+str(kPlanet.semiMajorAxis[2])+")"
								else: planetString += "; No Semi-major Axis data"
								if kPlanet.period[0]:
									planetString += "; 1.0/Period: "+str(kPlanet.period[0])
									if kPlanet.period[1] or kPlanet.period[2]: planetString += " (error -"+str(kPlanet.period[1])+" +"+str(kPlanet.period[2])+")"
								else: planetString += "; No Period data"
								missingDataLabels[missingObjectCount].setText(tab+planetString)
								missingDataLabels[missingObjectCount].setVisible(True)
								missingObjectCount += 1
								print "Missing "+planetString
							
							if kPlanet.description:
								objectDescriptionLabels[objectDescriptionCount].setText(tab+kPlanet.description)
								objectDescriptionLabels[objectDescriptionCount].setVisible(True)
								objectDescriptionCount += 1
				
				elif jChild[0] == 2:
					jStar = allStars[jChild[1]]
					if not jStar.radius[0]:
						starString = "Star: "+jStar.name[0]
						if jStar.spectralType[0]: starString += "; Spectral Type: "+jStar.spectralType[0]
						else: starString += "; Unknown Spectral Type"
						starString += "; No Radius / Mass data"
						missingDataLabels[missingObjectCount].setText(tab+starString)
						missingDataLabels[missingObjectCount].setVisible(True)
						missingObjectCount += 1
						print "Missing "+starString
					
					if not jStar.habitableInner or not jStar.habitableOuter or not maxPlanetOrbit:
						missingDataLabels[missingObjectCount].setText(tab+tab+jStar.name[0]+"'s habitiable zone, insufficient data for Luminosity and look-up table methods")
						missingDataLabels[missingObjectCount].setVisible(True)
						missingObjectCount += 1
						print "Missing Star: "+jStar.name[0]+"'s habitable zone"
					
					for kChild in jStar.children:
						kPlanet = allPlanets[kChild[1]]
						if not kPlanet.period[0] or not kPlanet.semiMajorAxis[0]:
							planetString = "Planet: "+kPlanet.name[0]
							if kPlanet.radius[0]:
								planetString += "; Radius: "+str(kPlanet.radius[0])
								if kPlanet.radius[1] or kPlanet.radius[2]: planetString += " (error -"+str(kPlanet.radius[1])+" +"+str(kPlanet.radius[2])+")"
							else: planetString += "; No Radius / Mass data"
							if kPlanet.semiMajorAxis[0]:
								planetString += "; Semi-major Axis: "+str(kPlanet.semiMajorAxis[0])
								if kPlanet.semiMajorAxis[1] or kPlanet.semiMajorAxis[2]: planetString += " (error -"+str(kPlanet.semiMajorAxis[1])+" +"+str(kPlanet.semiMajorAxis[2])+")"
							else: planetString += "; No Semi-major Axis data"
							if kPlanet.period[0]:
								planetString += "; 1.0/Period: "+str(kPlanet.period[0])
								if kPlanet.period[1] or kPlanet.period[2]: planetString += " (error -"+str(kPlanet.period[1])+" +"+str(kPlanet.period[2])+")"
							else: planetString += "; No Period data"
							missingDataLabels[missingObjectCount].setText(tab+tab+planetString)
							missingDataLabels[missingObjectCount].setVisible(True)
							missingObjectCount += 1
							print "Missing "+planetString
						
						if kPlanet.description:
							objectDescriptionLabels[objectDescriptionCount].setText(tab+kPlanet.description)
							objectDescriptionLabels[objectDescriptionCount].setVisible(True)
							objectDescriptionCount += 1
				
				elif jChild[0] == 3:
					jPlanet = allPlanets[jChild[1]]
					if not jPlanet.period[0] or not jPlanet.semiMajorAxis[0]:
						print "Missing Planet: "+jPlanet.name[0]
						planetString = "Planet: "+jPlanet.name[0]
						if jPlanet.radius[0]:
							planetString += "; Radius: "+str(jPlanet.radius[0])
							if jPlanet.radius[1] or jPlanet.radius[2]: planetString += " (error -"+str(jPlanet.radius[1])+" +"+str(jPlanet.radius[2])+")"
						else: planetString += "; No Radius / Mass data"
						if jPlanet.semiMajorAxis[0]:
							planetString += "; Semi-major Axis: "+str(jPlanet.semiMajorAxis[0])
							if jPlanet.semiMajorAxis[1] or jPlanet.semiMajorAxis[2]: planetString += " (error -"+str(jPlanet.semiMajorAxis[1])+" +"+str(jPlanet.semiMajorAxis[2])+")"
						else: planetString += "; No Semi-major Axis data"
						if jPlanet.period[0]:
							planetString += "; 1.0/Period: "+str(jPlanet.period[0])
							if jPlanet.period[1] or jPlanet.period[2]: planetString += " (error -"+str(jPlanet.period[1])+" +"+str(jPlanet.period[2])+")"
						else: planetString += "; No Period data"
						missingDataLabels[missingObjectCount].setText(tab+planetString)
						missingDataLabels[missingObjectCount].setVisible(True)
						missingObjectCount += 1
					
					if jPlanet.description:
						objectDescriptionLabels[objectDescriptionCount].setText(tab+jPlanet.description)
						objectDescriptionLabels[objectDescriptionCount].setVisible(True)
						objectDescriptionCount += 1
		
		elif iChild[0] == 2:
			iStar = allStars[iChild[1]]
			if not iStar.radius[0]:
				starString = "Star: "+iStar.name[0]
				if iStar.spectralType[0]: starString += "; Spectral Type: "+iStar.spectralType[0]
				else: starString += "; Unknown Spectral Type"
				starString += "; No Radius / Mass data"
				missingDataLabels[missingObjectCount].setText(tab+starString)
				missingDataLabels[missingObjectCount].setVisible(True)
				missingObjectCount += 1
				print "Missing "+starString
			
			if not iStar.habitableInner or not iStar.habitableOuter or not maxPlanetOrbit:
				missingDataLabels[missingObjectCount].setText(tab+tab+iStar.name[0]+"'s habitiable zone, insufficient data for Luminosity and look-up table methods")
				missingDataLabels[missingObjectCount].setVisible(True)
				missingObjectCount += 1
				print "Missing Star: "+iStar.name[0]+"'s habitable zone"
			
			for jChild in iStar.children:
				jPlanet = allPlanets[jChild[1]]
				if not jPlanet.semiMajorAxis[0] or not jPlanet.period[0]:
					planetString = "Planet: "+jPlanet.name[0]
					if jPlanet.radius[0]:
						planetString += "; Radius: "+str(jPlanet.radius[0])
						if jPlanet.radius[1] or jPlanet.radius[2]: planetString += " (error -"+str(jPlanet.radius[1])+" +"+str(jPlanet.radius[2])+")"
					else: planetString += "; No Radius / Mass data"
					if jPlanet.semiMajorAxis[0]:
						planetString += "; Semi-major Axis: "+str(jPlanet.semiMajorAxis[0])
						if jPlanet.semiMajorAxis[1] or jPlanet.semiMajorAxis[2]: planetString += " (error -"+str(jPlanet.semiMajorAxis[1])+" +"+str(jPlanet.semiMajorAxis[2])+")"
					else: planetString += "; No Semi-major Axis data"
					if jPlanet.period[0]:
						planetString += "; 1.0/Period: "+str(jPlanet.period[0])
						if jPlanet.period[1] or jPlanet.period[2]: planetString += " (error -"+str(jPlanet.period[1])+" +"+str(jPlanet.period[2])+")"
					else: planetString += "; No Period data"
					missingDataLabels[missingObjectCount].setText(tab+tab+planetString)
					missingDataLabels[missingObjectCount].setVisible(True)
					missingObjectCount += 1
					print "Missing "+planetString
				
				if jPlanet.description:
					objectDescriptionLabels[objectDescriptionCount].setText(tab+jPlanet.description)
					objectDescriptionLabels[objectDescriptionCount].setVisible(True)
					objectDescriptionCount += 1
		
		elif iChild[0] == 3:
			iPlanet = allPlanets[iChild[1]]
			if not iPlanet.period[0] or not iPlanet.semiMajorAxis[0]:
				planetString = "Planet: "+iPlanet.name[0]
				if iPlanet.radius[0]:
					planetString += "; Radius: "+str(iPlanet.radius[0])
					if iPlanet.radius[1] or iPlanet.radius[2]: planetString += " (error -"+str(iPlanet.radius[1])+" +"+str(iPlanet.radius[2])+")"
				else: planetString += "; No Radius / Mass data"
				if iPlanet.semiMajorAxis[0]:
					planetString += "; Semi-major Axis: "+str(iPlanet.semiMajorAxis[0])
					if iPlanet.semiMajorAxis[1] or iPlanet.semiMajorAxis[2]: planetString += " (error -"+str(iPlanet.semiMajorAxis[1])+" +"+str(iPlanet.semiMajorAxis[2])+")"
				else: planetString += "; No Semi-major Axis data"
				if iPlanet.period[0]:
					planetString += "; 1.0/Period: "+str(iPlanet.period[0])
					if iPlanet.period[1] or iPlanet.period[2]: planetString += " (error -"+str(iPlanet.period[1])+" +"+str(iPlanet.period[2])+")"
				else: planetString += "; No Period data"
				missingDataLabels[missingObjectCount].setText(tab+planetString)
				missingDataLabels[missingObjectCount].setVisible(True)
				missingObjectCount += 1
				print "Missing "+planetString
			
			if iPlanet.description:
				objectDescriptionLabels[objectDescriptionCount].setText(tab+iPlanet.description)
				objectDescriptionLabels[objectDescriptionCount].setVisible(True)
				objectDescriptionCount += 1
	
	# Manage Object Description Container
	for i in range(objectDescriptionCount, 10): objectDescriptionLabels[i].setVisible(False)
	if objectDescriptionCount == 1: objectDescriptions.setVisible(False)
	else: objectDescriptions.setVisible(True)
	
	# Manage Missing Data container
	for i in range(missingObjectCount, maxMultChildren): missingDataLabels[i].setVisible(False)
	if missingObjectCount == 1: missingData.setVisible(False)
	else: missingData.setVisible(True)
	
	# The fastest planet will make a complete orbit once per second
	maxPeriod /= twoPI
	
	print "Max period = "+str(maxPeriod)
	print "Max Binary orbit = "+str(maxBinaryOrbit)
	print "Max Star radius = "+str(maxStarRadius)
	print "Max Planet radius = "+str(maxPlanetRadius)
	print "Max Planet orbit = "+str(maxPlanetOrbit)
	
	print "Set "+inName+" as selected System instance"

# Initialize "Sun" system geometry as visible
setSystem("Sun")

# -----------------------------------------
# Multiples (CAVE2 screens are 1366 x 768)
# -----------------------------------------
# Keep track of what page each multiple is at
maxMultPageIndex = len(allSystems)//24
multPageIndex = [0, maxMultPageIndex//2]

# Load Image references
circleO = loadImage("images/circleO.png")
circleB = loadImage("images/circleB.png")
circleA = loadImage("images/circleA.png")
circleF = loadImage("images/circleF.png")
circleG = loadImage("images/circleG.png")
circleK = loadImage("images/circleK.png")
circleM = loadImage("images/circleM.png")
circleL = loadImage("images/circleL.png")
circleT = loadImage("images/circleT.png")
circleY = loadImage("images/circleY.png")
circleGreen = loadImage("images/circleGreen.png")
circleGray = loadImage("images/circleGray.png")
habitableBlock = loadImage("images/habitableBlock.jpg")

def setMultiples(inSide, inStart):
	endIndex = min(inStart+24, len(allSystems))
	for i in range(inStart, endIndex):
		objectIndex = i-inStart
		if not objectIndex: continue # Don't overwrite "Sun" System
		
		systemString = "System: "+allSystems[i].name[0]
		additionalNames = ""
		for j in range(1, len(allSystems[i].name)): additionalNames += allSystems[i].name[j]+", "
		if additionalNames: systemString += " ("+additionalNames[:-2]+")"
		multNameLabels[inSide][objectIndex][0].setText(systemString)
		
		systemString = ""
		if allSystems[i].distance[0]: systemString += "Distance: "+str(allSystems[i].distance[0])+" parsecs"
		if allSystems[i].distance[1] or allSystems[i].distance[2]: systemString += " (error -"+str(allSystems[i].distance[1])+" +"+str(allSystems[i].distance[2])+")"
		if allSystems[i].rightAscension and allSystems[i].declination:
			ascensionDeclinationString = "Right Ascension: "+allSystems[i].rightAscension+"; Declination: "+allSystems[i].declination
			if systemString: systemString += "; "+ascensionDeclinationString
			else: systemString = ascensionDeclinationString
		if not systemString: systemString = ""
		multNameLabels[inSide][objectIndex][1].setText(systemString)
		labelCount = 2
		
		objectHierarchy = []
		imageCount = 0
		
		for jChild in allSystems[i].children:
			if jChild[0] == 1:
				jBinary = allBinaries[jChild[1]]
				multNameLabels[inSide][objectIndex][labelCount].setText(tab+"Binary:")
				labelCount += 1
				
				for kChild in jBinary.children:
					if kChild[0] == 1:
						kBinary = allBinaries[kChild[1]]
						multNameLabels[inSide][objectIndex][labelCount].setText(tab+tab+"Binary:")
						labelCount += 1
						
						for lChild in kBinary.children:
							if lChild[0] == 1: continue # Ignore for now
							elif lChild[0] == 2:
								lStar = allStars[lChild[1]]
								starString = tab+tab+tab+"Star: "+lStar.name[0]
								if lStar.spectralType[0]: starString += "; Spectral Type: "+lStar.spectralType[0]
								multNameLabels[inSide][objectIndex][labelCount].setText(starString)
								labelCount += 1
								
								if lStar.spectralType[0]:
									spectralClass = lStar.spectralType[0][0]
									if spectralClass == "O": multObjectImages[inSide][objectIndex][imageCount].setData(circleO)
									elif spectralClass == "B": multObjectImages[inSide][objectIndex][imageCount].setData(circleB)
									elif spectralClass == "A": multObjectImages[inSide][objectIndex][imageCount].setData(circleA)
									elif spectralClass == "F": multObjectImages[inSide][objectIndex][imageCount].setData(circleF)
									elif spectralClass == "G": multObjectImages[inSide][objectIndex][imageCount].setData(circleG)
									elif spectralClass == "K": multObjectImages[inSide][objectIndex][imageCount].setData(circleK)
									elif spectralClass == "M": multObjectImages[inSide][objectIndex][imageCount].setData(circleM)
									elif spectralClass == "L": multObjectImages[inSide][objectIndex][imageCount].setData(circleL)
									elif spectralClass == "T": multObjectImages[inSide][objectIndex][imageCount].setData(circleT)
									elif spectralClass == "Y": multObjectImages[inSide][objectIndex][imageCount].setData(circleY)
									else: multObjectImages[inSide][objectIndex][imageCount].setData(circleGreen)
								else: multObjectImages[inSide][objectIndex][imageCount].setData(circleGreen)
								multObjectImages[inSide][objectIndex][imageCount].setVisible(True)
								objectHierarchy.append([])
								objectHierarchy[-1].append([imageCount, lStar])
								imageCount += 1
								
								for mChild in lStar.children:
									mPlanet = allPlanets[mChild[1]]
									planetString = tab+tab+tab+tab+"Planet: "+mPlanet.name[0]
									if mPlanet.discoveryMethod: planetString += "; Discovery Method: "+mPlanet.discoveryMethod
									if mPlanet.semiMajorAxis[0]: planetString += "; Semi-major Axis: "+str(mPlanet.semiMajorAxis[0])
									if mPlanet.semiMajorAxis[1] or mPlanet.semiMajorAxis[2]: planetString += " (error -"+str(mPlanet.semiMajorAxis[1])+" +"+str(mPlanet.semiMajorAxis[2])+")"
									multNameLabels[inSide][objectIndex][labelCount].setText(planetString)
									labelCount += 1
									
									multObjectImages[inSide][objectIndex][imageCount].setData(circleGray)
									multObjectImages[inSide][objectIndex][imageCount].setVisible(True)
									objectHierarchy[-1].append([imageCount, mPlanet])
									imageCount += 1
							
							elif lChild[0] == 3:
								lPlanet = allPlanets[lChild[1]]
								planetString = tab+tab+tab+"Planet: "+lPlanet.name[0]
								if lPlanet.discoveryMethod: planetString += "; Discovery Method: "+lPlanet.discoveryMethod
								multNameLabels[inSide][objectIndex][labelCount].setText(planetString)
								labelCount += 1
					
					elif kChild[0] == 2:
						kStar = allStars[kChild[1]]
						starString = tab+tab+"Star: "+kStar.name[0]
						if kStar.spectralType[0]: starString += "; Spectral Type: "+kStar.spectralType[0]
						multNameLabels[inSide][objectIndex][labelCount].setText(starString)
						labelCount += 1
						
						if kStar.spectralType[0]:
							spectralClass = kStar.spectralType[0][0]
							if spectralClass == "O": multObjectImages[inSide][objectIndex][imageCount].setData(circleO)
							elif spectralClass == "B": multObjectImages[inSide][objectIndex][imageCount].setData(circleB)
							elif spectralClass == "A": multObjectImages[inSide][objectIndex][imageCount].setData(circleA)
							elif spectralClass == "F": multObjectImages[inSide][objectIndex][imageCount].setData(circleF)
							elif spectralClass == "G": multObjectImages[inSide][objectIndex][imageCount].setData(circleG)
							elif spectralClass == "K": multObjectImages[inSide][objectIndex][imageCount].setData(circleK)
							elif spectralClass == "M": multObjectImages[inSide][objectIndex][imageCount].setData(circleM)
							elif spectralClass == "L": multObjectImages[inSide][objectIndex][imageCount].setData(circleL)
							elif spectralClass == "T": multObjectImages[inSide][objectIndex][imageCount].setData(circleT)
							elif spectralClass == "Y": multObjectImages[inSide][objectIndex][imageCount].setData(circleY)
							else: multObjectImages[inSide][objectIndex][imageCount].setData(circleGreen)
						else: multObjectImages[inSide][objectIndex][imageCount].setData(circleGreen)
						multObjectImages[inSide][objectIndex][imageCount].setVisible(True)
						objectHierarchy.append([])
						objectHierarchy[-1].append([imageCount, kStar])
						imageCount += 1
						
						for lChild in kStar.children:
							lPlanet = allPlanets[lChild[1]]
							planetString = tab+tab+tab+tab+"Planet: "+lPlanet.name[0]
							if lPlanet.discoveryMethod: planetString += "; Discovery Method: "+lPlanet.discoveryMethod
							if lPlanet.semiMajorAxis[0]: planetString += "; Semi-major Axis: "+str(lPlanet.semiMajorAxis[0])
							if lPlanet.semiMajorAxis[1] or lPlanet.semiMajorAxis[2]: planetString += " (error -"+str(lPlanet.semiMajorAxis[1])+" +"+str(lPlanet.semiMajorAxis[2])+")"
							multNameLabels[inSide][objectIndex][labelCount].setText(planetString)
							labelCount += 1
							
							multObjectImages[inSide][objectIndex][imageCount].setData(circleGray)
							multObjectImages[inSide][objectIndex][imageCount].setVisible(True)
							objectHierarchy[-1].append([imageCount, lPlanet])
							imageCount += 1
					
					elif kChild[0] == 3:
						kPlanet = allPlanets[kChild[1]]
						planetString = tab+tab+"Planet: "+kPlanet.name[0]
						if kPlanet.discoveryMethod: planetString += "; Discovery Method: "+kPlanet.discoveryMethod
						multNameLabels[inSide][objectIndex][labelCount].setText(planetString)
						labelCount += 1
			
			elif jChild[0] == 2:
				jStar = allStars[jChild[1]]
				starString = tab+"Star: "+jStar.name[0]
				if jStar.spectralType[0]: starString += "; Spectral Type: "+jStar.spectralType[0]
				multNameLabels[inSide][objectIndex][labelCount].setText(starString)
				labelCount += 1
				
				if jStar.spectralType[0]:
					spectralClass = jStar.spectralType[0][0]
					if spectralClass == "O": multObjectImages[inSide][objectIndex][imageCount].setData(circleO)
					elif spectralClass == "B": multObjectImages[inSide][objectIndex][imageCount].setData(circleB)
					elif spectralClass == "A": multObjectImages[inSide][objectIndex][imageCount].setData(circleA)
					elif spectralClass == "F": multObjectImages[inSide][objectIndex][imageCount].setData(circleF)
					elif spectralClass == "G": multObjectImages[inSide][objectIndex][imageCount].setData(circleG)
					elif spectralClass == "K": multObjectImages[inSide][objectIndex][imageCount].setData(circleK)
					elif spectralClass == "M": multObjectImages[inSide][objectIndex][imageCount].setData(circleM)
					elif spectralClass == "L": multObjectImages[inSide][objectIndex][imageCount].setData(circleL)
					elif spectralClass == "T": multObjectImages[inSide][objectIndex][imageCount].setData(circleT)
					elif spectralClass == "Y": multObjectImages[inSide][objectIndex][imageCount].setData(circleY)
					else: multObjectImages[inSide][objectIndex][imageCount].setData(circleGreen)
				else: multObjectImages[inSide][objectIndex][imageCount].setData(circleGreen)
				multObjectImages[inSide][objectIndex][imageCount].setVisible(True)
				objectHierarchy.append([])
				objectHierarchy[-1].append([imageCount, jStar])
				imageCount += 1
				
				for kChild in jStar.children:
					kPlanet = allPlanets[kChild[1]]
					planetString = tab+tab+"Planet: "+kPlanet.name[0]
					if kPlanet.discoveryMethod: planetString += "; Discovery Method: "+kPlanet.discoveryMethod
					if kPlanet.semiMajorAxis[0]: planetString += "; Semi-major Axis: "+str(kPlanet.semiMajorAxis[0])
					if kPlanet.semiMajorAxis[1] or kPlanet.semiMajorAxis[2]: planetString += " (error -"+str(kPlanet.semiMajorAxis[1])+" +"+str(kPlanet.semiMajorAxis[2])+")"
					multNameLabels[inSide][objectIndex][labelCount].setText(planetString)
					labelCount += 1
					
					multObjectImages[inSide][objectIndex][imageCount].setData(circleGray)
					multObjectImages[inSide][objectIndex][imageCount].setVisible(True)
					objectHierarchy[-1].append([imageCount, kPlanet])
					imageCount += 1
			
			elif jChild[0] == 3:
				jPlanet = allPlanets[jChild[1]]
				planetString = tab+"Planet: "+jPlanet.name[0]
				if jPlanet.discoveryMethod: planetString += "; Discovery Method: "+jPlanet.discoveryMethod
				multNameLabels[inSide][objectIndex][labelCount].setText(planetString)
				labelCount += 1
		
		# Clear out multiple children that were not needed
		for j in range(labelCount, maxMultChildren): multNameLabels[inSide][objectIndex][j].setText("")
		for j in range(imageCount, maxMultChildren): multObjectImages[inSide][objectIndex][j].setVisible(False)
		
		# Position, set width, set height, and set scale of all images
		starCount = len(objectHierarchy)
		if starCount:
			starLength = 768//starCount
			remainingSpace = 1366-starLength
			for j in range(starCount):
				starImage = multObjectImages[inSide][objectIndex][objectHierarchy[j][0][0]]
				starObject = objectHierarchy[j][0][1]
				
				starImage.setPosition(Vector2(remainingSpace, j*starLength))
				starImage.setWidth(starLength)
				starImage.setHeight(starLength)
				starImage.setScale(0.875)
				
				jObjects = len(objectHierarchy[j])
				planetLength = remainingSpace
				if jObjects-1: planetLength /= jObjects-1
				planetLength = min(planetLength, starLength)
				
				leftPosition = 0
				habitalbleLeft = 0x7fffffff
				for k in range(1, jObjects):
					planetImage = multObjectImages[inSide][objectIndex][objectHierarchy[j][k][0]]
					planetObject = objectHierarchy[j][k][1]
					
					planetImage.setPosition(Vector2(remainingSpace-planetLength*k, j*starLength+(starLength-planetLength)*0.5))
					planetImage.setWidth(planetLength)
					planetImage.setHeight(planetLength)
					planetImage.setScale(0.75)
					
					if planetObject.semiMajorAxis[0] and starObject.habitableInner and starObject.habitableOuter:
						if planetObject.semiMajorAxis[0] > starObject.habitableInner and planetObject.semiMajorAxis[0] < starObject.habitableOuter:
							leftPosition = remainingSpace-planetLength*k
							habitalbleLeft = leftPosition if leftPosition < habitalbleLeft else habitalbleLeft
				
				if habitalbleLeft == 0x7fffffff: multHabitableImages[inSide][objectIndex][j].setVisible(False)
				else:
					multHabitableImages[inSide][objectIndex][j].setPosition(Vector2(leftPosition, j*starLength))
					multHabitableImages[inSide][objectIndex][j].setWidth(planetLength)
					multHabitableImages[inSide][objectIndex][j].setHeight(starLength)
					multHabitableImages[inSide][objectIndex][j].setVisible(True)
			
			# Remove habitable images with an index >= starCount
			for j in range(starCount, 3): multHabitableImages[inSide][objectIndex][j].setVisible(False)
		
		# Remove habitable images for orphan planets
		else:
			for j in range(3): multHabitableImages[inSide][objectIndex][j].setVisible(False)
	
	# Clear out final multiples if on the last page
	for i in range(endIndex, inStart+24):
		objectIndex = i-inStart
		for j in range(maxMultChildren):
			multNameLabels[inSide][objectIndex][j].setText("")
			multObjectImages[inSide][objectIndex][j].setVisible(False)
		for j in range(3): multHabitableImages[inSide][objectIndex][j].setVisible(False)

# 0 is left and 1 is right
mult = [[], []]
multNameLabels = [[], []]
multObjectImages = [[], []]
multHabitableImages = [[], []]

# Create left and right multiple Container instances and their children
for i in range(2):
	for j in range(24):
		jx = j%6
		jy = j//6
		
		mult[i].append(wf.createContainer("Multiple:"+str(i)+":"+str(j), uiRoot, ContainerLayout.LayoutFree))
		mult[i][j].setStyle("fill:#20202080")
		mult[i][j].setSize(Vector2(1366, 768))
		mult[i][j].setPosition(Vector2(i*16392+jx*1366, jy*768))
		if clippingOn: mult[i][j].setClippingEnabled(True) # So far not needed
		mult[i][j].setAutosize(False)
		
		multNameLabels[i].append([])
		multObjectImages[i].append([])
		for k in range(maxMultChildren):
			multNameLabels[i][j].append(wf.createLabel("MultipleNameLabel:"+str(i)+":"+str(j)+":"+str(k), mult[i][j], " ")) # Apparently empty string is bad
			multNameLabels[i][j][k].setStyle("color:#ff8080; font:fonts/arial.ttf 24")
			multNameLabels[i][j][k].setPosition(Vector2(8, 12+k*32))
			multNameLabels[i][j][k].setLayer(WidgetLayer.Front)
			
			multObjectImages[i][j].append(wf.createImage("MultipleObjectImage:"+str(i)+":"+str(j)+":"+str(k), mult[i][j]))
			multObjectImages[i][j][k].setLayer(WidgetLayer.Middle)
		
		multHabitableImages[i].append([])
		for k in range(3):
			multHabitableImages[i][j].append(wf.createImage("MultipleHabitableImage:"+str(i)+":"+str(j)+":"+str(k), mult[i][j]))
			multHabitableImages[i][j][k].setData(habitableBlock)
			multHabitableImages[i][j][k].setLayer(WidgetLayer.Back)

# Assign all but the first index for both left and right
setMultiples(0, multPageIndex[0]*24)
setMultiples(1, multPageIndex[1]*24)

# Manually assign the first index for both left and right as "Sun"
for i in range(2):
	multNameLabels[i][0][0].setText("System: Sun")
	multNameLabels[i][0][1].setText(tab+"Star: Sun: G2V")
	multNameLabels[i][0][2].setText(tab+tab+"Planet: Mercury")
	multNameLabels[i][0][3].setText(tab+tab+"Planet: Venus")
	multNameLabels[i][0][4].setText(tab+tab+"Planet: Earth")
	multNameLabels[i][0][5].setText(tab+tab+"Planet: Mars")
	multNameLabels[i][0][6].setText(tab+tab+"Planet: Jupiter")
	multNameLabels[i][0][7].setText(tab+tab+"Planet: Saturn")
	multNameLabels[i][0][8].setText(tab+tab+"Planet: Uranus")
	multNameLabels[i][0][9].setText(tab+tab+"Planet: Neptune")
	multNameLabels[i][0][10].setText(tab+tab+"Planet: Pluto")
	
	multObjectImages[i][0][0].setData(circleG)
	multObjectImages[i][0][1].setData(circleGray)
	multObjectImages[i][0][2].setData(circleGray)
	multObjectImages[i][0][3].setData(circleGray)
	multObjectImages[i][0][4].setData(circleGray)
	multObjectImages[i][0][5].setData(circleGray)
	multObjectImages[i][0][6].setData(circleGray)
	multObjectImages[i][0][7].setData(circleGray)
	multObjectImages[i][0][8].setData(circleGray)
	multObjectImages[i][0][9].setData(circleGray)
	
	multObjectImages[i][0][0].setPosition(Vector2(598, 0))
	planetLength = 598/9
	for j in range(1, 10):
		multObjectImages[i][0][j].setPosition(Vector2(598-planetLength*j, 384-planetLength*0.5))
		multObjectImages[i][0][j].setWidth(planetLength)
		multObjectImages[i][0][j].setHeight(planetLength)
		multObjectImages[i][0][j].setScale(0.75)
	
	multHabitableImages[i][0][0].setPosition(Vector2(598-planetLength*3, 0))
	multHabitableImages[i][0][0].setWidth(planetLength)
	for j in range(1, 3): multHabitableImages[i][0][j].setVisible(False)

# --------------------------------
# Animate current System instance
# --------------------------------
def onUpdate(frame, t, dt):
	global periodTheta
	
	sphereCount = 0
	lineSetCount = 0
	lightCount = 0
	periodCount = 0
	
	for iChild in allSystems[currSystem].children:
		if iChild[0] == 1:
			iBinary = allBinaries[iChild[1]]
			iBinaryTheta = 0.0
			if iBinary.period[0] and iBinary.semiMajorAxis[0]:
				periodTheta[periodCount] += dt*iBinary.period[0]/maxPeriod*userPeriod
				iBinaryTheta = periodTheta[periodCount]
				periodCount += 1
			
			for jChild in iBinary.children:
				if jChild[0] == 1:
					ix = 0.0
					iy = 0.0
					if iBinary.semiMajorAxis[0]:
						ix = math.cos(iBinaryTheta)*iBinary.semiMajorAxis[0]/maxBinaryOrbit*userBinaryOrbit
						iy = math.sin(iBinaryTheta)*iBinary.semiMinorAxis/maxBinaryOrbit*userBinaryOrbit
						iBinaryTheta += math.pi
					
					jBinary = allBinaries[jChild[1]]
					jBinaryTheta = 0.0
					if jBinary.period[0] and jBinary.semiMajorAxis[0]:
						periodTheta[periodCount] += dt*jBinary.period[0]/maxPeriod*userPeriod
						jBinaryTheta = periodTheta[periodCount]
						periodCount += 1
					
					for kChild in jBinary.children:
						if kChild[0] == 1: continue # Ignore for now...
						elif kChild[0] == 2:
							jx = ix
							jy = iy
							if jBinary.semiMajorAxis[0]:
								jx = ix+math.cos(jBinaryTheta)*jBinary.semiMajorAxis[0]/maxBinaryOrbit*userBinaryOrbit
								jy = iy+math.sin(jBinaryTheta)*jBinary.semiMinorAxis/maxBinaryOrbit*userBinaryOrbit
								jBinaryTheta += math.pi
							
							kStar = allStars[kChild[1]]
							if kStar.radius[0]:
								allSpheres[sphereCount].setPosition(Vector3(jx, jy, 0.0))
								sphereCount += 1
							
							if kStar.habitableInner and kStar.habitableOuter and maxPlanetOrbit:
								hRadius = (kStar.habitableOuter+kStar.habitableInner)*0.5
								linearScale = hRadius/maxPlanetOrbit*userPlanetOrbit
								
								allLineSets[lineSetCount].setPosition(Vector3(jx, jy, 0.0))
								theta = 0.0
								previousPoint = Vector3(linearScale, 0.0, 0.0)
								for segs in range(segmentsPerLineSet):
									theta += thetaInc
									currPoint = Vector3(cos(theta)*linearScale, sin(theta)*linearScale, 0.0)
									allLineSetCaps[lineSetCount][segs].setPosition(Vector3(jx, jy, 0.0)+previousPoint)
									previousPoint = currPoint
								lineSetCount += 1
							
							allLights[lightCount].setPosition(Vector3(jx, jy, 0.0))
							lightCount += 1
							
							for lChild in kStar.children:
								lPlanet = allPlanets[lChild[1]]
								if lPlanet.semiMajorAxis[0] and lPlanet.period[0]:
									periodTheta[periodCount] += dt*lPlanet.period[0]/maxPeriod*userPeriod
									lx = jx+math.cos(periodTheta[periodCount])*lPlanet.semiMajorAxis[0]/maxPlanetOrbit*userPlanetOrbit
									ly = jy+math.sin(periodTheta[periodCount])*lPlanet.semiMinorAxis/maxPlanetOrbit*userPlanetOrbit
									periodCount += 1
									
									allSpheres[sphereCount].setPosition(Vector3(lx, ly, 0.0))
									sphereCount += 1
									
									allLineSets[lineSetCount].setPosition(Vector3(jx, jy, 0.0))
									theta = 0.0
									previousPoint = Vector3(lPlanet.semiMajorAxis[0]/maxPlanetOrbit*userPlanetOrbit, 0.0, 0.0)
									for segs in range(segmentsPerLineSet):
										theta += thetaInc
										currPoint = Vector3(cos(theta)*lPlanet.semiMajorAxis[0]/maxPlanetOrbit*userPlanetOrbit, sin(theta)*lPlanet.semiMinorAxis/maxPlanetOrbit*userPlanetOrbit, 0.0)
										allLineSetCaps[lineSetCount][segs].setPosition(Vector3(jx, jy, 0.0)+previousPoint)
										previousPoint = currPoint
									lineSetCount += 1
						
						elif kChild[0] == 3:
							kPlanet = allPlanets[kChild[1]]
							if kPlanet.period[0] and kPlanet.semiMajorAxis[0]:
								periodTheta[periodCount] += dt*kPlanet.period[0]/maxPeriod*userPeriod
								kx = ix+math.cos(periodTheta[periodCount])*kPlanet.semiMajorAxis[0]/maxPlanetOrbit*userPlanetOrbit
								ky = iy+math.sin(periodTheta[periodCount])*kPlanet.semiMinorAxis/maxPlanetOrbit*userPlanetOrbit
								periodCount += 1
								
								allSpheres[sphereCount].setPosition(Vector3(kx, ky, 0.0))
								sphereCount += 1
								
								allLineSets[lineSetCount].setPosition(Vector3(ix, iy, 0.0))
								theta = 0.0
								previousPoint = Vector3(kPlanet.semiMajorAxis[0]/maxPlanetOrbit*userPlanetOrbit, 0.0, 0.0)
								for segs in range(segmentsPerLineSet):
									theta += thetaInc
									currPoint = Vector3(cos(theta)*kPlanet.semiMajorAxis[0]/maxPlanetOrbit*userPlanetOrbit, sin(theta)*kPlanet.semiMinorAxis/maxPlanetOrbit*userPlanetOrbit, 0.0)
									allLineSetCaps[lineSetCount][segs].setPosition(Vector3(jx, jy, 0.0)+previousPoint)
									previousPoint = currPoint
								lineSetCount += 1
				
				elif jChild[0] == 2:
					ix = 0.0
					iy = 0.0
					if iBinary.semiMajorAxis[0]:
						ix = math.cos(iBinaryTheta)*iBinary.semiMajorAxis[0]/maxBinaryOrbit*userBinaryOrbit
						iy = math.sin(iBinaryTheta)*iBinary.semiMinorAxis/maxBinaryOrbit*userBinaryOrbit
						iBinaryTheta += math.pi
					
					jStar = allStars[jChild[1]]
					if jStar.radius[0]:
						allSpheres[sphereCount].setPosition(Vector3(ix, iy, 0.0))
						sphereCount += 1
					
					if jStar.habitableInner and jStar.habitableOuter and maxPlanetOrbit:
						hRadius = (jStar.habitableOuter+jStar.habitableInner)*0.5
						linearScale = hRadius/maxPlanetOrbit*userPlanetOrbit
						
						allLineSets[lineSetCount].setPosition(Vector3(ix, iy, 0.0))
						theta = 0.0
						previousPoint = Vector3(linearScale, 0.0, 0.0)
						for segs in range(segmentsPerLineSet):
							theta += thetaInc
							currPoint = Vector3(cos(theta)*linearScale, sin(theta)*linearScale, 0.0)
							allLineSetCaps[lineSetCount][segs].setPosition(Vector3(ix, iy, 0.0)+previousPoint)
							previousPoint = currPoint
						lineSetCount += 1
					
					allLights[lightCount].setPosition(Vector3(ix, iy, 0.0))
					lightCount += 1
					
					for kChild in jStar.children:
						kPlanet = allPlanets[kChild[1]]
						if kPlanet.semiMajorAxis[0] and kPlanet.period[0]:
							periodTheta[periodCount] += dt*kPlanet.period[0]/maxPeriod*userPeriod
							kx = ix+math.cos(periodTheta[periodCount])*kPlanet.semiMajorAxis[0]/maxPlanetOrbit*userPlanetOrbit
							ky = iy+math.sin(periodTheta[periodCount])*kPlanet.semiMinorAxis/maxPlanetOrbit*userPlanetOrbit
							periodCount += 1
							
							allSpheres[sphereCount].setPosition(Vector3(kx, ky, 0.0))
							sphereCount += 1
							
							allLineSets[lineSetCount].setPosition(Vector3(ix, iy, 0.0))
							theta = 0.0
							previousPoint = Vector3(kPlanet.semiMajorAxis[0]/maxPlanetOrbit*userPlanetOrbit, 0.0, 0.0)
							for segs in range(segmentsPerLineSet):
								theta += thetaInc
								currPoint = Vector3(cos(theta)*kPlanet.semiMajorAxis[0]/maxPlanetOrbit*userPlanetOrbit, sin(theta)*kPlanet.semiMinorAxis/maxPlanetOrbit*userPlanetOrbit, 0.0)
								allLineSetCaps[lineSetCount][segs].setPosition(Vector3(ix, iy, 0.0)+previousPoint)
								previousPoint = currPoint
							lineSetCount += 1
				
				elif jChild[0] == 3:
					jPlanet = allPlanets[jChild[1]]
					if jPlanet.period[0] and jPlanet.semiMajorAxis[0]:
						periodTheta[periodCount] += dt*jPlanet.period[0]/maxPeriod*userPeriod
						jx = math.cos(periodTheta[periodCount])*jPlanet.semiMajorAxis[0]/maxPlanetOrbit*userPlanetOrbit
						jy = math.sin(periodTheta[periodCount])*jPlanet.semiMinorAxis/maxPlanetOrbit*userPlanetOrbit
						periodCount += 1
						
						allSpheres[sphereCount].setPosition(Vector3(jx, jy, 0.0))
						sphereCount += 1
						lineSetCount += 1
		
		elif iChild[0] == 2:
			iStar = allStars[iChild[1]]
			if iStar.radius[0]: sphereCount += 1
			lightCount += 1
			
			for jChild in iStar.children:
				jPlanet = allPlanets[jChild[1]]
				if jPlanet.semiMajorAxis[0] and jPlanet.period[0]:
					periodTheta[periodCount] += dt*jPlanet.period[0]/maxPeriod*userPeriod
					jx = math.cos(periodTheta[periodCount])*jPlanet.semiMajorAxis[0]/maxPlanetOrbit*userPlanetOrbit
					jy = math.sin(periodTheta[periodCount])*jPlanet.semiMinorAxis/maxPlanetOrbit*userPlanetOrbit
					periodCount += 1
					
					allSpheres[sphereCount].setPosition(Vector3(jx, jy, 0.0))
					sphereCount += 1
					lineSetCount += 1
		
		elif iChild[0] == 3:
			iPlanet = allPlanets[iChild[1]]
			if iPlanet.period[0] and iPlanet.semiMajorAxis[0]: sphereCount += 1
setUpdateFunction(onUpdate)

# --------------------------------
# Add "Select System" to the menu
# --------------------------------
# Make a sorted list of all System instance names (systemDict cannot be sorted)
systemNames = sorted(systemDict.keys())

# "Select \"Sun\" System" is always an option
menuSelectSystem = mainMenu.getMainMenu().addSubMenu("Select System")
menuSelectSystem.addButton("Select \"Sun\" System", "setSystem('Sun')")

# System instance names will be distributed into menus
selectSystemMenus = []
selectSystemButtons = []
selectSystemPages = []
snLen = len(systemNames)
numSSMenus = int(ceil(snLen/100.0)) # ...why does Python return a float for ceil()

def updateSelectSystemMenu(inIndex, inAlter):
	global selectSystemPages
	
	# Ensure the new page has a valid index
	newSSPage = selectSystemPages[inIndex]+inAlter
	if newSSPage >= 0 and newSSPage < 9: # 9 is ceil(100/12)
		selectSystemPages[inIndex] = newSSPage
		
		# Update selectSystemButtons[inIndex]
		firstIndex = inIndex*100
		lastIndex = min(firstIndex+100, snLen-1)
		firstIndex += newSSPage*12
		for i in range(12):
			iIndex = firstIndex+i
			if iIndex < lastIndex:
				selectSystemButtons[inIndex][i].setText("Select \""+systemNames[iIndex]+"\" System")
				selectSystemButtons[inIndex][i].setCommand("setSystem('"+systemNames[iIndex]+"')")
			else:
				selectSystemButtons[inIndex][i].setText("-")
				selectSystemButtons[inIndex][i].setCommand("print '...not a System instance name'")
	print "Viewing page "+str(selectSystemPages[inIndex])+" of submenu "+str(inIndex)

# Assign selectSystemMenus[i] snLen/100 menus and selectSystemButtons[i] the first 12 appropriate System instance buttons
for i in range(numSSMenus):
	firstIndex = i*100
	lastIndex = min(firstIndex+99, snLen-1)
	selectSystemMenus.append(menuSelectSystem.addSubMenu("Submenu \""+systemNames[firstIndex]+"\" to \""+systemNames[lastIndex]+"\""))
	selectSystemButtons.append([])
	selectSystemPages.append(0)
	
	# Previous is always an option in selectSystemMenus[i]
	selectSystemMenus[-1].addButton("Previous", "updateSelectSystemMenu("+str(i)+", -1)")
	
	# Append the first 12 appropriate System instance name buttons to selectSystemMenus[i]
	lastButton = firstIndex+12
	for j in range(firstIndex, lastButton):
		selectSystemButtons[-1].append(selectSystemMenus[-1].addButton("Select \""+systemNames[j]+"\" System", "setSystem('"+systemNames[j]+"')"))
	
	# Next is always an option in selectSystemMenus[i]
	selectSystemMenus[-1].addButton("Next", "updateSelectSystemMenu("+str(i)+", 1)")

# -----------------------------------
# Add "Select Multiples" to the Menu
# -----------------------------------
def changeMultiplePage(inSide, inAlter):
	global multPageIndex
	
	if multPageIndex[inSide]+inAlter >= 0 and multPageIndex[inSide]+inAlter <= maxMultPageIndex:
		multPageIndex[inSide] += inAlter
		if not inSide: menuSelectMultiplesLeftLabel.setText("Left Multiple Page: "+str(multPageIndex[0])+" of "+str(maxMultPageIndex))
		else: menuSelectMultiplesRightLabel.setText("Right Multiple Page: "+str(multPageIndex[1])+" of "+str(maxMultPageIndex))
		setMultiples(inSide, multPageIndex[inSide]*24)
	else: print "...invalid multiple page"

menuSelectMultiples = mainMenu.getMainMenu().addSubMenu("Select Multiples")

menuSelectMultiplesLeftLabel = menuSelectMultiples.addLabel("Left Multiple Page: "+str(multPageIndex[0])+" of "+str(maxMultPageIndex))
menuSelectMultiples.addButton("Left Previous Page", "changeMultiplePage(0, -1)")
menuSelectMultiples.addButton("Left Next Page", "changeMultiplePage(0, 1)")

menuSelectMultiples.addLabel("")

menuSelectMultiplesRightLabel = menuSelectMultiples.addLabel("Right Multiple Page: "+str(multPageIndex[1])+" of "+str(maxMultPageIndex))
menuSelectMultiples.addButton("Right Previous Page", "changeMultiplePage(1, -1)")
menuSelectMultiples.addButton("Right Next Page", "changeMultiplePage(1, 1)")

# -------------------------------
# Add "Scale System" to the menu
# -------------------------------
menuScaleSystem = mainMenu.getMainMenu().addSubMenu("Scale System")

def scalePeriod(inAlter):
	global userPeriod
	userPeriod *= inAlter
	menuScalePeriodLabel.setText("Current Scale: "+str(userPeriod))
menuScalePeriod = menuScaleSystem.addSubMenu("Scale Time")
menuScalePeriodLabel = menuScalePeriod.addLabel("Current Scale: "+str(userPeriod))
menuScalePeriod.addButton("Time * 1.25", "scalePeriod("+str(1.25)+")")
menuScalePeriod.addButton("Time * 0.8", "scalePeriod("+str(0.8)+")")

def scaleBinaryOrbit(inAlter):
	global userBinaryOrbit, userPlanetOrbit
	userBinaryOrbit *= inAlter
	menuScaleBinaryOrbitLabel.setText("Current Scale: "+str(userBinaryOrbit))
	if constantDistance:
		userPlanetOrbit = userBinaryOrbit
		menuScalePlanetOrbitLabel.setText("Current Scale: "+str(userPlanetOrbit))
	scaleSystemGeometry()
menuScaleBinaryOrbit = menuScaleSystem.addSubMenu("Scale Binary Orbit")
menuScaleBinaryOrbitLabel = menuScaleBinaryOrbit.addLabel("Current Scale: "+str(userBinaryOrbit))
menuScaleBinaryOrbit.addButton("Binary Orbit * 1.25", "scaleBinaryOrbit("+str(1.25)+")")
menuScaleBinaryOrbit.addButton("Binary Orbit * 0.8", "scaleBinaryOrbit("+str(0.8)+")")

def scaleStarRadius(inAlter):
	global userStarRadius
	userStarRadius *= inAlter
	menuScaleStarRadiusLabel.setText("Current Scale: "+str(userStarRadius))
	scaleSystemGeometry()
menuScaleStarRadius = menuScaleSystem.addSubMenu("Scale Star Radii")
menuScaleStarRadiusLabel = menuScaleStarRadius.addLabel("Current Scale: "+str(userStarRadius))
menuScaleStarRadius.addButton("Star Radii * 1.25", "scaleStarRadius("+str(1.25)+")")
menuScaleStarRadius.addButton("Star Radii * 0.8", "scaleStarRadius("+str(0.8)+")")

def scalePlanetRadius(inAlter):
	global userPlanetRadius
	userPlanetRadius *= inAlter
	menuScalePlanetRadiusLabel.setText("Current Scale: "+str(userPlanetRadius))
	scaleSystemGeometry()
menuScalePlanetRadius = menuScaleSystem.addSubMenu("Scale Planet Radii")
menuScalePlanetRadiusLabel = menuScalePlanetRadius.addLabel("Current Scale: "+str(userPlanetRadius))
menuScalePlanetRadius.addButton("Planet Radii * 1.25", "scalePlanetRadius("+str(1.25)+")")
menuScalePlanetRadius.addButton("Planet Radii * 0.8", "scalePlanetRadius("+str(0.8)+")")

def scalePlanetOrbit(inAlter):
	global userBinaryOrbit, userPlanetOrbit
	userPlanetOrbit *= inAlter
	menuScalePlanetOrbitLabel.setText("Current Scale: "+str(userPlanetOrbit))
	if constantDistance:
		userBinaryOrbit = userPlanetOrbit
		menuScaleBinaryOrbitLabel.setText("Current Scale: "+str(userBinaryOrbit))
	scaleSystemGeometry()
menuScalePlanetOrbit = menuScaleSystem.addSubMenu("Scale Planet Orbit")
menuScalePlanetOrbitLabel = menuScalePlanetOrbit.addLabel("Current Scale: "+str(userPlanetOrbit))
menuScalePlanetOrbit.addButton("Planet Orbit * 1.25", "scalePlanetOrbit("+str(1.25)+")")
menuScalePlanetOrbit.addButton("Planet Orbit * 0.8", "scalePlanetOrbit("+str(0.8)+")")

def toggleConstantDistance():
	global constantDistance
	constantDistance = not constantDistance
	print "Constant Distance: "+str(constantDistance)
	resetUserScales()
	setSystem(allSystems[currSystem].name[0])
constantDistanceButton = menuScaleSystem.addButton("Constant Distance", "toggleConstantDistance()")
constantDistanceButton.getButton().setCheckable(True)
constantDistanceButton.getButton().setChecked(False)

def resetUserScales():
	global userPeriod, userBinaryOrbit, userStarRadius, userPlanetRadius, userPlanetOrbit
	userPeriod = 1.0
	userBinaryOrbit = 10.0
	userStarRadius = 1.0
	userPlanetRadius = 1.0
	userPlanetOrbit = 10.0
	menuScalePeriodLabel.setText("Current Scale: "+str(userPeriod))
	menuScaleBinaryOrbitLabel.setText("Current Scale: "+str(userBinaryOrbit))
	menuScaleStarRadiusLabel.setText("Current Scale: "+str(userStarRadius))
	menuScalePlanetRadiusLabel.setText("Current Scale: "+str(userPlanetRadius))
	menuScalePlanetOrbitLabel.setText("Current Scale: "+str(userPlanetOrbit))
	cam.setPosition(Vector3(0.0, 0.0, 15.0))
	cam.setPitchYawRoll(Vector3(0.0, 0.0, 0.0))
menuScaleSystem.addButton("Reset Scales", "resetUserScales(); scaleSystemGeometry()")

menuScaleSystem.addButton("Reset Camera", "cam.setPosition(Vector3(0.0, 0.0, 15.0)); cam.setPitchYawRoll(Vector3(0.0, 0.0, 0.0))")

# -------------------------------------------------------
# Demon magic that converts a Wand ray to 2D coordinates
# -------------------------------------------------------
coorCalc = CoordinateCalculator()

def onEvent():
	e = getEvent()
	if (e.getServiceType() == ServiceType.Wand):
		r = getRayFromEvent(e)
		if (r[0]):
			o = e.getOrientation() # Get something that has more dimensions than I understand
			p = e.getPosition() # Get the position of the wand
			q = Quaternion(o.w, o.x, o.y, o.z) # Do something with quaternions
			v = q*Vector3(0.0, 0.0, -1.0) # Somehow this is allowed in Python
			
			coorCalc.set_position(p.x, p.y, p.z)
			coorCalc.set_orientation(v.x, v.y, v.z)
			coorCalc.calculate()
			
			ccx = coorCalc.get_x()
			ccy = coorCalc.get_y()
			
			if ccx >= 0.0 and ccy >= 0.0:
				pixelVec = Vector2(ccx*24588, ccy*3072)
				for i in range(2):
					for j in range(24):
						if mult[i][j].hitTest(pixelVec):
							mult[i][j].setStyle("fill:#ff000080")
							if (e.isButtonDown(EventFlags.ButtonUp)):
								if not j: setSystem("Sun")
								else:
									print "Selected "+allSystems[multPageIndex[i]+j].name[0]
									setSystem(allSystems[multPageIndex[i]*24+j].name[0])
						else: mult[i][j].setStyle("fill:#20202080")
setEventFunction(onEvent)
